// component_dialog.cpp : implementation file
//

#include <math.h>
#include <stdio.h>

#include "..\stdafx.h"
#include "..\component.h"
#include "component_dialog.h"
#include "..\linux_port\main_extern.h"
#include "..\linux_port\first_fit.h"
#include "..\linux_port\first_fit-tiebreaker.h"
#include "..\linux_port\load_balance.h"
#include "..\linux_port\comm_min-first_fit.h"
#include "..\linux_port\comm_min-load_balance.h"
#include "..\linux_port\greedy_b_and_b.h"
#include "..\linux_port\optimal_mapping.h"
#include "warning.h"
#include "proc_details.h"
#include "choose_algo_dialog.h"
#include "../Random System Generator/gen_sys.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//runs algorithm given and checks to see if any utilization is returned that could not
//be scheduled - if so an error message is displayed
static void RunAndCheckAlgo(double (*algo)());

//This is by far the most confusing function prototype I have ever written:
//RunAndCheckAlgo takes three parameters:
//Parameter1 - algo: this is a function pointer to the algorithm that is to be run
//					 it takes the second and third parameters as its first and second
//					 parameters respectively
//Parameter2 - tiebreaker: this is a function object that is called by the algo function that
//					 takes in a list of CommMinNode* that have previously been determined 
//					 to be in a tie (see parameter3) and this function will return one 
//					 CommMinNode* to break this tie
//Parameter3 - IsATie: this is another function object that is called by the algo function
//					 that has 3 parameters: the first is a CommMinNode* that is being checked
//					 to see if it is in a tie with any other nodes that are in parameter2 which
//					 is a set of nodes.  If there are any nodes that are determined to be in
//					 a tie with the given node these nodes should be put in the 3rd CommMinNode*
//					 list
//The function runs the algo given with its parameters and checks if any extra utilization
//is returned - if so an error message is displayed
static void RunAndCheckAlgo(double (*algo)(CommMinNode* (*)(CommMinNode*, list<CommMinNode*>&), bool (*)(CommMinNode*, multiset<CommMinNode*, UtilizationGreaterThan>, list<CommMinNode*>&)), 
							CommMinNode* (*tiebreaker)(CommMinNode*, list<CommMinNode*>&), 
							bool (*IsATie)(CommMinNode*, std::multiset<CommMinNode*, UtilizationGreaterThan>, std::list<CommMinNode*>&));

static void UseRandomSys();
static void UseGMESys();

//externals
extern bool all_blocks_to_procs;
extern bool disable_comm_min;
extern bool use_gme_hw;
extern bool optimal_comm_min;
extern bool disable_mem_algos;
extern user_preferences prefs;
extern Graph rand_grp;		//graph and system generated randomly
extern System rand_sys;
extern Graph GME_grp;		//graph and system from GME model
extern System GME_sys;	
extern Graph grp;			//graph and system used in algorithms - either from random or GME systems and models
extern System sys;

/////////////////////////////////////////////////////////////////////////////
// component_dialog dialog


component_dialog::component_dialog(CWnd* pParent /*=NULL*/)
	: CDialog(component_dialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(component_dialog)
	//}}AFX_DATA_INIT
}

void component_dialog::initialize(bool priority_period_on)
{
	fo_period_based_rb		= static_cast<CButton*>(GetDlgItem(IDC_RADIO1));
	fo_priority_based_rb	= static_cast<CButton*>(GetDlgItem(IDC_RADIO2));
	fo_first_fit_rb			= static_cast<CButton*>(GetDlgItem(IDC_RADIO3));
	fo_load_balanced_rb		= static_cast<CButton*>(GetDlgItem(IDC_RADIO4));
	fo_comm_min_rb			= static_cast<CButton*>(GetDlgItem(IDC_RADIO9));
	so_period_based_rb		= static_cast<CButton*>(GetDlgItem(IDC_RADIO5));
	so_priority_based_rb	= static_cast<CButton*>(GetDlgItem(IDC_RADIO6));
	so_first_fit_rb			= static_cast<CButton*>(GetDlgItem(IDC_RADIOFIRSTFIT2));
	so_load_balanced_rb		= static_cast<CButton*>(GetDlgItem(IDC_LOADBALANCE2));
	so_comm_min_rb			= static_cast<CButton*>(GetDlgItem(IDC_COMM_MIN));
	view_rand_sys			= static_cast<CButton*>(GetDlgItem(IDC_VIEW_RAND_SYS));
	use_rand_sys_rb			= static_cast<CButton*>(GetDlgItem(IDC_USE_RAND_SYS));
	use_gme_rb				= static_cast<CButton*>(GetDlgItem(IDC_USE_GME));
	multiple_constraints_cb = static_cast<CButton*>(GetDlgItem(IDC_MULTIPLE_CONSTRAINTS_CHECK));

	//disable the first order priority and period mapping radio buttons 
	if(!priority_period_on)
	{
		fo_period_based_rb->EnableWindow(FALSE);		// Period Based
		fo_priority_based_rb->EnableWindow(FALSE);		// Priority Based
	}

	if(disable_comm_min)
		fo_comm_min_rb->EnableWindow(FALSE);			// Comm Min

	if(disable_mem_algos)
		multiple_constraints_cb->EnableWindow(FALSE);

	CComboBox* c_box = static_cast<CComboBox*>(GetDlgItem(IDC_COMBO1));
	int i = 0;

	GME_sys.Order_Procs_By_Name();

	//populate the combo box with the names of the processors
	for(list<Processor*>::iterator it = GME_sys.Proc_Begin(); it != GME_sys.Proc_End(); ++it)
		c_box->InsertString(i++, (*it)->Property()->Get_Name());

	//no mapping policies have been chosen 
	policy.fo_mp = NONE;
	policy.so_mp = NONE;

	//default is to run first fit
	fo_first_fit_rb->SetCheck(BST_CHECKED);
	OnFirstFitMap();

	//set the radio button to use the GME model (instead of a random model) to be true
	use_gme_rb->SetCheck(BST_CHECKED);

	//default is to use the system in the GME model
	UseGMESys();
}

void component_dialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	initialize(all_blocks_to_procs);
	//{{AFX_DATA_MAP(component_dialog)
	DDX_Control(pDX, IDC_EDIT2, m_edit);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(component_dialog, CDialog)
	//{{AFX_MSG_MAP(component_dialog)
	ON_BN_CLICKED(IDC_RADIO1, OnSubsystemMap)
	ON_BN_CLICKED(IDC_RADIO2, OnPriorityMap)
	ON_BN_CLICKED(IDC_BUTTON1, OnDoMap)
	ON_BN_CLICKED(IDC_RADIO3, OnFirstFitMap)
	ON_BN_CLICKED(IDC_RADIO4, OnLoadBalanceMap)
	ON_BN_CLICKED(IDC_RADIO9, OnCommMinMap)
	ON_BN_CLICKED(IDC_RADIO5, OnSubsystemMapSecond)
	ON_BN_CLICKED(IDC_RADIO6, OnPriorityMapSecond)
	ON_BN_CLICKED(IDC_RADIOFIRSTFIT2, OnFirstFitMapSecond)
	ON_BN_CLICKED(IDC_LOADBALANCE2, OnLoadBalanceMapSecond)
	ON_BN_CLICKED(IDC_COMM_MIN, OnCommMinMapSecond)
	ON_BN_CLICKED(IDC_BUTTON2, OnResetPreferences)
	ON_BN_CLICKED(IDC_BUTTON3, OnButton3)
	ON_BN_CLICKED(IDC_BUTTON4, OnGenRandSys)
	ON_BN_CLICKED(IDC_USE_RAND_SYS, OnUseRandSys)
	ON_BN_CLICKED(IDC_USE_GME, OnUseGmeSys)
	ON_BN_CLICKED(IDC_VIEW_RAND_SYS, OnViewRandSys)
//	ON_BN_CLICKED(IDC_MULTIPLE_CONSTRAINTS_CHECK, OnMultipleConstraintsCheck)
	ON_BN_CLICKED(IDC_MULTIPLE_CONSTRAINTS_RB, OnMultipleConstraints)
	ON_BN_CLICKED(IDC_OPTIMAL_RB, OnOptimal)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// component_dialog message handlers

/* First Order Mapping */
void component_dialog::OnSubsystemMap() 
{
	policy.fo_mp=PERIOD_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, FALSE);
}

void component_dialog::OnPriorityMap() 
{
	policy.fo_mp=PRIORITY_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, FALSE);
}

void component_dialog::OnFirstFitMap() 
{
	policy.fo_mp=FIRSTFIT_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, TRUE);
}

void component_dialog::OnLoadBalanceMap() 
{
	policy.fo_mp=LOAD_BALANCE_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, FALSE);
}

void component_dialog::OnCommMinMap() 
{
	policy.fo_mp=COMM_MIN_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, TRUE, TRUE, FALSE);
}

void component_dialog::OnMultipleConstraints() 
{
	policy.fo_mp = MULTIPLE_CONSTRAINTS_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, FALSE);	
}

void component_dialog::OnOptimal() 
{
	policy.fo_mp = OPTIMAL_MAPPING;
	EnableDisableSecondOrder(FALSE, FALSE, FALSE, FALSE, FALSE);
}


/* Second Order Mapping */
void component_dialog::OnSubsystemMapSecond() 
{
	if(policy.so_mp == PERIOD_MAPPING)
	{
		policy.so_mp=NONE;
		so_period_based_rb->SetCheck(BST_UNCHECKED);
	}
	else
		policy.so_mp=PERIOD_MAPPING;
}

void component_dialog::OnPriorityMapSecond() 
{
	if(policy.so_mp == PRIORITY_MAPPING)
	{
		policy.so_mp=NONE;
		so_priority_based_rb->SetCheck(BST_UNCHECKED);
	}
	else
		policy.so_mp=PRIORITY_MAPPING;
}

void component_dialog::OnFirstFitMapSecond() 
{
	if(policy.so_mp == FIRSTFIT_MAPPING)
	{
		policy.so_mp=NONE;
		so_first_fit_rb->SetCheck(BST_UNCHECKED);
	}
	else
		policy.so_mp=FIRSTFIT_MAPPING;
}

void component_dialog::OnLoadBalanceMapSecond() 
{
	if(policy.so_mp == LOAD_BALANCE_MAPPING)
	{
		policy.so_mp=NONE;
		so_load_balanced_rb->SetCheck(BST_UNCHECKED);
	}
	else
		policy.so_mp=LOAD_BALANCE_MAPPING;
}

void component_dialog::OnCommMinMapSecond() 
{	
	if(policy.so_mp == COMM_MIN_MAPPING)
	{
		policy.so_mp=NONE;
		so_comm_min_rb->SetCheck(BST_UNCHECKED);
	}
	else
		policy.so_mp=COMM_MIN_MAPPING;
}

void component_dialog::OnDoMap() 
{
	mbuff.Empty();
	
    switch(policy.fo_mp)
    {
        case PRIORITY_MAPPING: 
            Do_Priority_Mapping();
            break;
        case PERIOD_MAPPING:
            Do_Period_Mapping();
            break;
		case FIRSTFIT_MAPPING:
			if(policy.so_mp == COMM_MIN_MAPPING)
				RunAndCheckAlgo(Do_FirstFit_Mapping, CommMinTiebreaker, CommMinIsATie);
			else
				RunAndCheckAlgo(Do_FirstFit_Mapping);
			break;
		case LOAD_BALANCE_MAPPING:
			RunAndCheckAlgo(Do_LoadBalance_Mapping);
            break;
		case COMM_MIN_MAPPING:
			if(policy.so_mp == LOAD_BALANCE_MAPPING)
			{
				ChooseAlgoDlg dlg;
				if(dlg.DoModal() == IDOK)
					RunAndCheckAlgo(Do_CommMinLoadBalance_Mapping);
				else
					return;
			}
			else
			{
				int rval = cm_algo_choose.DoModal();
				if(rval == OPTIMAL)
					optimal_comm_min = true;
				else
					optimal_comm_min = false;

				if(rval != ALGO_CANCEL)
					RunAndCheckAlgo(Do_CommMinFirstFit_Mapping);
			}
			break;
		case MULTIPLE_CONSTRAINTS_MAPPING:
			RunAndCheckAlgo(run_greedy_b_and_b);
			break;
		case OPTIMAL_MAPPING:
			RunAndCheckAlgo(do_optimal_mapping);
			break;
		case NONE:
			mbuff += "Please select a first order mapping algorithm";
			break;
		default:
			mbuff += "Unknown Mapping Algorithm";
			break;
    }

	m_edit.SetWindowText(mbuff);
	return;
}

void component_dialog::OnResetPreferences() 
{
	Set_Default_Preferences();
}

void component_dialog::OnButton3() 
{
	CComboBox* c_box = static_cast<CComboBox*>(GetDlgItem(IDC_COMBO1));
	int index = c_box->GetCurSel();

	if(index == -1)
	{
		AfxMessageBox("Please choose a processor to examine.", MB_OK | MB_ICONSTOP);
		return;
	}

	int i = 0;
	for(list<Processor*>::iterator it = sys.Proc_Begin(); it != sys.Proc_End(); ++it)
		if(i++ == index) break;

	if(it != sys.Proc_End())
		show_grouping(*it);
}

void component_dialog::EnableDisableSecondOrder(int period, int priority, int first_fit, int load_balance, int comm_min)
{
	so_period_based_rb->EnableWindow(period);
	so_priority_based_rb->EnableWindow(priority);
	so_first_fit_rb->EnableWindow(first_fit);
	so_load_balanced_rb->EnableWindow(load_balance);
	so_comm_min_rb->EnableWindow(comm_min);

	if(!period)
	{
		if(policy.so_mp == PERIOD_MAPPING)
			policy.so_mp = NONE;
		so_period_based_rb->SetCheck(BST_UNCHECKED);
	}
	if(!priority)
	{
		if(policy.so_mp == PRIORITY_MAPPING)
			policy.so_mp = NONE;
		so_priority_based_rb->SetCheck(BST_UNCHECKED);
	}
	if(!first_fit)
	{
		if(policy.so_mp == FIRSTFIT_MAPPING)
			policy.so_mp = NONE;
		so_first_fit_rb->SetCheck(BST_UNCHECKED);
	}
	if(!load_balance)
	{
		if(policy.so_mp == LOAD_BALANCE_MAPPING)
			policy.so_mp = NONE;
		so_load_balanced_rb->SetCheck(BST_UNCHECKED);
	}
	if(!comm_min)
	{
		if(policy.so_mp == COMM_MIN_MAPPING)
			policy.so_mp = NONE;
		so_comm_min_rb->SetCheck(BST_UNCHECKED);
	}
}

void component_dialog::OnGenRandSys() 
{
	if(rand_sys_generator.DoModal() == IDOK && rand_grp.Get_Number_Of_Nodes())
	{
		UseRandomSys();

		//enable the view random system button and the use GME and use random sys radio buttons
		view_rand_sys->EnableWindow(TRUE);
		use_rand_sys_rb->EnableWindow(TRUE);
		use_gme_rb->EnableWindow(TRUE);

		//select the use random system radio button and unselect the use GME radio button
		use_rand_sys_rb->SetCheck(BST_CHECKED);
		use_gme_rb->SetCheck(BST_UNCHECKED);
	}
}

void component_dialog::OnUseRandSys() 
{
	UseRandomSys();	
}

void component_dialog::OnUseGmeSys() 
{
	UseGMESys();	
}

void component_dialog::OnViewRandSys() 
{
	if(use_gme_hw)
	{
		if(GME_sys.GetDescription() == "")
			GenerateDescription(&GME_sys);
	}
	else
	{
		if(rand_sys.GetDescription() == "")
			GenerateDescription(&rand_sys);
	}

	if(rand_grp.GetDescription() == "")
		GenerateDescription(&rand_grp);

	if(use_gme_hw)
		view_system.SetText(GME_sys.GetDescription() + rand_grp.GetDescription());
	else
		view_system.SetText(rand_sys.GetDescription() + rand_grp.GetDescription());

	view_system.DoModal();
}
/*
void component_dialog::OnMultipleConstraintsCheck() 
{
	if(multiple_constraints_cb->GetCheck() == BST_CHECKED)
	{
		fo_period_based_rb->EnableWindow(  FALSE);
		fo_priority_based_rb->EnableWindow(FALSE);
		fo_first_fit_rb->EnableWindow(     FALSE);
		fo_load_balanced_rb->EnableWindow( FALSE);
		fo_comm_min_rb->EnableWindow(      FALSE);
		so_period_based_rb->EnableWindow(  FALSE);
		so_priority_based_rb->EnableWindow(FALSE);
		so_first_fit_rb->EnableWindow(     FALSE);
		so_load_balanced_rb->EnableWindow( FALSE);
		so_comm_min_rb->EnableWindow(      FALSE);

		policy.fo_mp = MULTIPLE_CONSTRAINTS_MAPPING;
	}
	else
	{
		fo_period_based_rb->SetCheck(BST_UNCHECKED);
		fo_priority_based_rb->SetCheck(BST_UNCHECKED);
		fo_first_fit_rb->SetCheck(BST_CHECKED);
		fo_load_balanced_rb->SetCheck(BST_UNCHECKED);
		fo_comm_min_rb->SetCheck(BST_UNCHECKED);

		fo_first_fit_rb->EnableWindow(TRUE);
		fo_load_balanced_rb->EnableWindow(TRUE);

		//disable the first order priority and period mapping radio buttons 
		if(all_blocks_to_procs)
		{
			fo_period_based_rb->EnableWindow(TRUE);
			fo_priority_based_rb->EnableWindow(TRUE);
		}

		if(!disable_comm_min)
			fo_comm_min_rb->EnableWindow(TRUE);
		OnFirstFitMap();
	}
}*/

void show_Mapping_Error(double extra_utilization)
{
	string msg;

	char extra_procs_10[10];	//for bound of 1.0
	char extra_procs_08[10];	//for bound of 0.8
	char extra_procs_06[10];	//for bound of 0.6
	char extra_procs_04[10];	//for bound of 0.4

	//multiply extra_utilization by the inverse of the bound
	sprintf(extra_procs_10, "%i", (int)ceil(extra_utilization));
	sprintf(extra_procs_08, "%i", (int)ceil(extra_utilization * 1.25));
	sprintf(extra_procs_06, "%i", (int)ceil(extra_utilization * 1.67));
	sprintf(extra_procs_04, "%i", (int)ceil(extra_utilization * 2.5));

	msg = string("There are not enough processors available to process all of the tasks.") +
		"  Please adjust the bounds of the existing CPU's or add more processors.\n\n" + 
		"For a bound of 1.0:\t " + extra_procs_10 + " extra processor(s) are needed\n" +
		"For a bound of 0.8:\t " + extra_procs_08 + " extra processor(s) are needed\n" +
		"For a bound of 0.6:\t " + extra_procs_06 + " extra processor(s) are needed\n" +
		"For a bound of 0.4:\t " + extra_procs_04 + " extra processor(s) are needed\n";

	AfxMessageBox(msg.c_str(), MB_OK | MB_ICONSTOP);
}

void RunAndCheckAlgo(double (*algo)())
{
	double r;
	r = algo();	//returns 0 if everything is ok, if not it returns
				//the utilization that could not be put on a processor

	if(r > 0)		//error
		show_Mapping_Error(r);
	else if(r < 0)
		AfxMessageBox("No valid solution was found.  However, a valid solution may exist.", MB_OK | MB_ICONSTOP);
}

void RunAndCheckAlgo(double (*algo)(CommMinNode* (*)(CommMinNode*, list<CommMinNode*>&), bool (*)(CommMinNode*, multiset<CommMinNode*, UtilizationGreaterThan>, list<CommMinNode*>&)), CommMinNode* (*tiebreaker)(CommMinNode*, list<CommMinNode*>&), bool (*IsATie)(CommMinNode*, std::multiset<CommMinNode*, UtilizationGreaterThan>, std::list<CommMinNode*>&))
{
	double r;

	r = algo(tiebreaker, IsATie);	//returns 0 if everything is ok, if not it returns
									//the utilization that could not be put on a processor

	if(r > 0)		//error
		show_Mapping_Error(r);
	else if(r < 0)
		AfxMessageBox("No valid solution was found.  However, a valid solution may exist.", MB_OK | MB_ICONSTOP);
}

void UseRandomSys()
{
	if(use_gme_hw)
		sys = GME_sys;
	else
		sys = rand_sys;

	grp = rand_grp;
}

void UseGMESys()
{
	sys = GME_sys;
	grp = GME_grp;
}