#include "..\stdafx.h"
#include "windows.h"
#include <stdio.h>
#include <map>
#include <algorithm>
#include <string>
#include "debug.h"
#include "main_extern.h"
#include "system.h"
#include "schedule.h"
#include "comm_min-first_fit.h"

using namespace std;

//externals
extern System sys;
extern __int64 num_steps;
extern double g_best_comm_cost;
extern double g_best_balanced_load_cost;
extern double g_best_first_fit_cost;

extern map<Processor*, double> proc_util_for_comm;
extern map<Processor*, double> proc_util_for_load_balance;
extern map<Processor*, double> proc_util_for_first_fit;

extern double load_balance_comm_cost;
extern double first_fit_comm_cost;

//delete later
extern double abs(double d);

typedef struct
{
    list<TaskType*> tlist;
} task_struct;

void Print_Task_List(list<TaskType *>& tl)
{
    list<TaskType*>::iterator pi;

    lout.Format("\r\nPrinting List..."); debug_print();
    for(pi=tl.begin();pi!=tl.end();pi++)
    {
        Print_Task(*pi);
    }
    return;
}

void Print_Task (TaskType *pt)
{
	if(pt->Property()->Get_ID()==0) return;

    float util=pt->Property()->Get_Exec_Time()/pt->Property()->Get_Period();
    TaskProperty* ptp=pt->Property();
    lout.Format("\r\nTask %s", ptp->Get_Name());debug_print();
    Processor *p=ptp->Get_Proc();
    if(p)
	{
        lout.Format(" On Processor %s",p->Property()->Get_Name());
		debug_print();
	}
    /*
	lout.Format("\r\n  Offset %.3f  Deadline %.3f Exe.(ms) %.3f Period (ms) %.3f Util. %.3f Resp.(ms) %.3f CSW %d",
            ptp->Get_Offset(),ptp->Get_Deadline(),
			ptp->Get_Exec_Time(), ptp->Get_Period(), util, ptp->Get_Response_Time(),
            ptp->Get_Context_Switches());
	debug_print();
    */
	lout.Format("\r\nDeadline %.3f Exe.(ms) %.3f Period (ms) %.3f Util. %.3f Resp.(ms) %.3f CSW %d Priority %d",
            ptp->Get_Deadline(),
			ptp->Get_Exec_Time(), ptp->Get_Period(), util, ptp->Get_Response_Time(),
            ptp->Get_Context_Switches(),ptp->Get_Priority());
    debug_print();
    return;
}

void Print_Task_Deadline (TaskType *pt)
{
    if(pt->Property()->Get_ID()==0) return;
    
    lout.Format("\r\n%-30.30s\t%.3f\t\t%.3f\t\t%.3f",pt->Property()->Get_Name(),
            pt->Property()->Get_Offset(),
            pt->Property()->Get_Exec_Time(),
            pt->Property()->Get_Deadline());
    debug_print();
    return;
}
    

void Print_Connection (ConnectionType *pc)
{
    TaskProperty* ftp=pc->Get_From()->Property();
    TaskProperty* ttp=pc->Get_To()->Property();

    float sft=ftp->Get_Deadline()-(ftp->Get_Offset()+ftp->Get_Exec_Time());
    float stt=ttp->Get_Deadline()-(ttp->Get_Offset()+ttp->Get_Exec_Time());
    float sc=ttp->Get_Offset()-(ftp->Get_Deadline()+pc->Property()->Get_Trans_Time());

    float f=(sft+stt+sc)/(ftp->Get_Exec_Time()+ttp->Get_Exec_Time()+
            pc->Property()->Get_Trans_Time());
    
    lout.Format("\r\nMessage from %s To %s",
            pc->Get_From()->Property()->Get_Name(),
            pc->Get_To()->Property()->Get_Name());
	debug_print();
    
    Processor* from=pc->Get_From()->Property()->Get_Proc();
    Processor* to=pc->Get_To()->Property()->Get_Proc();
    if((from)&&(to)&&(from!=to))
	{
        lout.Format(" On Network");
		debug_print();
	}
    
    lout.Format("\r\n    Offset %.3f Deadline %.3f Size(bytes) %d Trans.(ms) %.3f Resp.(ms) %.3f",
			pc->Get_From()->Property()->Get_Deadline(),pc->Get_To()->Property()->Get_Offset(),
            pc->Property()->Get_Message_Size(),
            pc->Property()->Get_Trans_Time(),pc->Property()->Get_Response_Time());
	debug_print();
    return;
}

void Print_Connection_Deadline (ConnectionType* pc)
{
    TaskProperty* ftp=pc->Get_From()->Property();
    TaskProperty* ttp=pc->Get_To()->Property();

    if(ftp->Get_ID()==0) return;
    CString name="("+ftp->Get_Name()+","+ttp->Get_Name()+")";
    lout.Format("\r\n%-30.30s\t%.3f\t\t%.3f\t\t%.3f",
            name,ftp->Get_Deadline(),
            pc->Property()->Get_Trans_Time(),ttp->Get_Offset());
    debug_print();
    return;
}

void Print_Graph (Graph& g)
{
    list<TaskType*>::iterator ip;

    for(ip=g.Node_Begin();ip!=g.Node_End();ip++)
    {
        TaskType* pt=*ip;

        Print_Task(pt);
    }
    return;
}

void Print_Graph_Deadline (Graph& g)
{
    list<TaskType*>::iterator ip;

    lout.Format("%-60.60s","Tasks");
    debug_print();
    lout.Format("\r\n-----------------------------------------------------------------------------------------------------------------------------------\r\n");
    debug_print();

    for(ip=g.Node_Begin();ip!=g.Node_End();ip++)
    {
        TaskType* pt=*ip;

        Print_Task_Deadline(pt);
    }
    return;
}

void Print_Connection (Graph& g)
{
    list<ConnectionType*>::iterator ic;
    
    for(ic=g.Edge_Begin();ic!=g.Edge_End();ic++)
    {
        ConnectionType* pc=*ic;

        Print_Connection(pc);
    }
}

void Print_Connection_Deadline (Graph& g)
{
    list<ConnectionType*>::iterator ic;

    lout.Format("\r\n\r\n%-60.60s","Messages");
    debug_print();
    lout.Format("\r\n-----------------------------------------------------------------------------------------------------------------------------------\r\n");
    debug_print();

    for(ic=g.Edge_Begin();ic!=g.Edge_End();ic++)
    {
        ConnectionType* pc=*ic;

        Print_Connection_Deadline(pc);
    }
}

void Print_Processor (Processor *p)
{
	ProcessorProperty* pp=p->Get_Property();
	float task_util=Get_Task_Utilization(p);
	float csw_util=Get_Context_Switch_Utilization(p);
	float sched_util=Get_Scheduling_Utilization(p);
	float timer_util=Get_Timer_Utilization(p);

	lout.Format("\r\nProcessor %s", pp->Get_Name());debug_print();
	lout.Format("\r\n\tTask Util. %.3f Sched. Util %.3f Timer Util %.3f CSW Util %.3f",
		task_util,sched_util,timer_util,csw_util);
	debug_print();
}


void Print_System (System& s)
{
	list<Processor*>::iterator pi;

	for(pi=s.Proc_Begin();pi!=s.Proc_End();pi++)
	{
		Print_Processor(*pi);
	}
}

void Print_Priority_Task_Map ()
{
    list<task_struct> tsl;
    list<TaskType*>::iterator ti;

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()!=pt) continue;
        list<TaskType*>* topol=pt->Property()->Topo();
        list<TaskType*>::iterator ti2;
        int prev=-1;
        task_struct ts;
        
        for(ti2=topol->begin();ti2!=topol->end();ti2++)
        {
            TaskType* pt2=*ti2;

            if(pt2->Property()->Get_Priority()!=prev)
            {
                if(!ts.tlist.empty()) tsl.push_back(ts);
                prev=pt2->Property()->Get_Priority();
                ts.tlist.clear();
                ts.tlist.push_back(pt2);
            }
            else
            {
                ts.tlist.push_front(pt2);
            }
        }
        if(!ts.tlist.empty()) tsl.push_back(ts);
    }

    list<task_struct>::iterator tsli;

    for(tsli=tsl.begin();tsli!=tsl.end();tsli++)
    {
        task_struct ts=*tsli;
        TaskType* pt;
        
        ti=ts.tlist.begin();
        pt=*ti;
        lout.Format("\r\n\r\nTask %s Priority %d Processor %s Includes :- ", pt->Property()->Get_Name(), pt->Property()->Get_Priority(), pt->Property()->Get_Alloc_Proc()->Property()->Get_Name());debug_print();
        for(;ti!=ts.tlist.end();ti++)
        {
            lout.Format("\r\n\t\t\t%s",(*ti)->Property()->Get_Name());debug_print();
        }
    }
    return;
}
            
void Print_Period_Task_Map (map<TaskType*,TaskType*>& tm)
{
    list<TaskType*> mapped,done;
    list<TaskType*>::iterator ti;
    map<TaskType*,TaskType*>::const_iterator mi,mi2;

    for(mi=tm.begin();mi!=tm.end();mi++)
    {
        TaskType* pt2=mi->second;

        ti=find(done.begin(), done.end(), pt2);
        if(ti!=done.end()) continue;
        done.push_front(pt2);
        mapped.clear();
        
        for(mi2=tm.begin();mi2!=tm.end();mi2++)
        {
            TaskType* pt=mi2->first;
            TaskType* pt3=mi2->second;
            
            if(pt3==pt2) mapped.push_front(pt);
        }
        
        lout.Format("\r\nTask %s Priority %d Processor %s Include :-", pt2->Property()->Get_Name(),pt2->Property()->Get_Priority(),pt2->Property()->Get_Alloc_Proc()->Property()->Get_Name()); debug_print();
        
        for(ti=mapped.begin();ti!=mapped.end();ti++)
        {
            TaskType* pt=*ti;

            list<TaskType*>* topol=pt->Property()->Topo();
            list<TaskType*>::iterator ti2;

            for(ti2=topol->begin();ti2!=topol->end();ti2++)
            {
                TaskType* pt3=*ti2;
                lout.Format("\r\n\t\t\t%s",pt3->Property()->Get_Name());debug_print();
            }
        }
    }
    return;
}

void print_FirstFit_Mapping()
{
	double comm_cost  = FindCommCost();
	double total_comm = FindTotalComm();
	double percentage;
	
	if(total_comm)
		percentage = comm_cost/total_comm*100;
	else
		percentage = 0;

	//print title
	lout.Format("-------------------------------------------------------------------------------------First Fit--------------------------------------------------------------------------------------------"); debug_print();
	lout.Format("\r\nSteps:\t%d", num_steps); debug_print();
	lout.Format("\r\n\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%", comm_cost, total_comm, percentage); debug_print();

	print_Utilization_Mapping();
}

void print_LoadBalance_Mapping()
{
	double comm_cost  = FindCommCost();
	double total_comm = FindTotalComm();
	double percentage;
	
	if(total_comm)
		percentage = comm_cost/total_comm*100;
	else
		percentage = 0;

	//print title
	lout.Format("------------------------------------------------------------------------------------- Load Balance --------------------------------------------------------------------------------------------"); debug_print();
	lout.Format("\r\nSteps:\t%d", num_steps); debug_print();
	lout.Format("\r\n\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%", comm_cost, total_comm, percentage); debug_print();

	print_Utilization_Mapping();
}

void print_CommMin_Mapping(double comm_cost)
{
	double total_comm = FindTotalComm();
	double unbalance_util = 0;
	double percentage;

	lout.Format("--------------------------------------------------------------------- Communication Minimization -----------------------------------------------------------------------------------------"); debug_print();

//delete later
		for(list<Processor*>::iterator proc_it1 = sys.Proc_Begin(); proc_it1 != sys.Proc_End(); ++proc_it1)
			for(list<Processor*>::iterator proc_it2 = proc_it1; proc_it2 != sys.Proc_End(); ++proc_it2)
			{
				double dif = (((*proc_it1)->Property()->Get_Bound() - (*proc_it1)->Property()->Get_Available_Space()) - 
				              ((*proc_it2)->Property()->Get_Bound() - (*proc_it2)->Property()->Get_Available_Space()));
				unbalance_util += abs(dif);
			}
//

	if(total_comm)
		percentage = comm_cost/total_comm*100;
	else
		percentage = 0;

	unsigned int high_order_bits = num_steps/4294967296;	//bits 63:32
	unsigned int low_order_bits  = num_steps%4294967296;	//bits 31:0
	
	if(high_order_bits)
		{lout.Format("\r\nSteps:\t%d%d", high_order_bits, low_order_bits); debug_print();}
	else
		{lout.Format("\r\nSteps:\t%d", low_order_bits); debug_print();}

	lout.Format("\r\n\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%", comm_cost, total_comm, percentage); debug_print();

	//delete later
	lout.Format("\r\nUnbalanced Utilization:\t%f", unbalance_util); debug_print();

	print_Utilization_Mapping();
}

void print_Utilization_Mapping()
{
	//go through all processors
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		lout.Format("\r\n\r\n\r\nProcessor:   %s\tTotal Utilization:   %f\tBound:   %f\r\n", 
			proc->Property()->Get_Name(), proc->Property()->Get_Bound() - proc->Property()->Get_Available_Space(), proc->Property()->Get_Bound()); debug_print();

		lout.Format("----------------------------------------------------------------------------------------------------------------------------------------"); debug_print();

		//go through each task that was mapped to this processor
		for(list<TaskType*>::iterator task_it = proc->Tasks_Begin(); task_it != proc->Tasks_End(); ++task_it)
		{
			TaskProperty* property = (*task_it)->Property();

			//create propper spacing between the name of the task and its utilization
			//numbers were derived through trial and error - feel free to change if
			//some other algorithm works better
			string s;
			s =  property->Get_Name() + '\t';
			if(s.length() < 9)
				s += string("\t") + string("\t") + string("\t");
			else if(s.length() < 20)
				s += string("\t") + string("\t");
			else if(s.length() < 30)
				s += string("\t");

			double utilization = property->Get_Exec_Time() / property->Get_Period();
			lout.Format("\r\n\t\t%sUtilization:   %f", s.c_str(), utilization); debug_print();
		}
	}
}

void print_MultipleConstraints_Mapping()
{
	double comm_cost  = FindCommCost();
	double total_comm = FindTotalComm();
	double percentage;
	
	if(total_comm)
		percentage = comm_cost/total_comm*100;
	else
		percentage = 0;

	//print title
	lout.Format("-------------------------------------------------------------------------------------Multiple Constraints--------------------------------------------------------------------------------------------"); debug_print();

	lout.Format("\r\n\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%", comm_cost, total_comm, percentage); debug_print();

	//go through all processors
	for(list<Processor*>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		lout.Format("\r\n\r\n\r\nProcessor:   %s\r\nTotal Utilization:   %f\tBound:   %f\tTotal Memory:   %f\tMemory Cost:   %d\r\n", 
			proc->Property()->Get_Name(), proc->Property()->Get_Bound() - proc->Property()->Get_Available_Space(), proc->Property()->Get_Bound(), proc->Property()->Get_GBB_Property()->GetMemUsed(), proc->Property()->Get_GBB_Property()->GetMemCost()); debug_print();

		lout.Format("----------------------------------------------------------------------------------------------------------------------------------------"); debug_print();

		//go through each task that was mapped to this processor
		for(list<TaskType*>::iterator task_it = proc->Tasks_Begin(); task_it != proc->Tasks_End(); ++task_it)
		{
			TaskProperty* property = (*task_it)->Property();

			//create propper spacing between the name of the task and its utilization
			//numbers were derived through trial and error - feel free to change if
			//some other algorithm works better
			string s;
			s =  property->Get_Name() + '\t';
			if(s.length() < 9)
				s += string("\t") + string("\t") + string("\t");
			else if(s.length() < 20)
				s += string("\t") + string("\t");
			else if(s.length() < 30)
				s += string("\t");

			double utilization = property->Get_Exec_Time() / property->Get_Period();
			lout.Format("\r\n\t\t%sUtilization:   %f", s.c_str(), utilization); debug_print();
		}
	}
}

void print_Optimal_Mapping()
{
	list<Processor*>::iterator proc_it;

	double total_comm = FindTotalComm();
	double ff_comm_percentage;
	double lb_comm_percentage;
	double cm_comm_percentage;
	
	if(total_comm)
	{
		ff_comm_percentage = first_fit_comm_cost/total_comm*100;
		lb_comm_percentage = load_balance_comm_cost/total_comm*100;
		cm_comm_percentage = g_best_comm_cost/total_comm*100;
	}
	else
	{
		ff_comm_percentage = 0;
		lb_comm_percentage = 0;
		cm_comm_percentage = 0;
	}

	unsigned int high_order_bits = num_steps/4294967296;	//bits 63:32
	unsigned int low_order_bits  = num_steps%4294967296;	//bits 31:0
	
	if(high_order_bits)
		{lout.Format("\r\nSteps:\t%d%d\r\n", high_order_bits, low_order_bits); debug_print();}
	else
		{lout.Format("\r\nSteps:\t%d\r\n", low_order_bits); debug_print();}

	//print first fit title
	lout.Format("-------------------------------------------------------------------------------------Optimum First Fit--------------------------------------------------------------------------------------------"); debug_print();
	lout.Format("\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%\r\n\r\n", first_fit_comm_cost, total_comm, ff_comm_percentage); debug_print();

	//go through all processors
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		lout.Format("Processor:   %s\tTotal Utilization:   %f\tBound:   %f\r\n", 
			proc->Property()->Get_Name(), proc_util_for_first_fit[proc], proc->Property()->Get_Bound()); debug_print();
	}
	
	lout.Format("__________________________________________________________________________________________________________________________________________________________________________________________________\r\n\r\n\r\n"); debug_print();


	//print load balance title
	lout.Format("------------------------------------------------------------------------------------Optimum Load Balance------------------------------------------------------------------------------------------"); debug_print();
	lout.Format("\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%\r\n\r\n", load_balance_comm_cost, total_comm, lb_comm_percentage); debug_print();		
	
	//go through all processors
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		lout.Format("Processor:   %s\tTotal Utilization:   %f\tBound:   %f\r\n", 
			proc->Property()->Get_Name(), proc_util_for_load_balance[proc], proc->Property()->Get_Bound()); debug_print();
	}

	lout.Format("__________________________________________________________________________________________________________________________________________________________________________________________________\r\n\r\n\r\n"); debug_print();


	//print communication minimization title
	lout.Format("------------------------------------------------------------------------------Optimum Communication Minimization----------------------------------------------------------------------------------"); debug_print();
	lout.Format("\r\nCommunication cost:\t%f\r\nTotal Communication:\t%f\r\nPercentage:\t\t%f%%\r\n\r\n", g_best_comm_cost, total_comm, cm_comm_percentage); debug_print();
	
	//go through all processors
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc = *proc_it;
		lout.Format("Processor:   %s\tTotal Utilization:   %f\tBound:   %f\r\n", 
			proc->Property()->Get_Name(), proc_util_for_comm[proc], proc->Property()->Get_Bound()); debug_print();
	}

	lout.Format("__________________________________________________________________________________________________________________________________________________________________________________________________\r\n\r\n\r\n"); debug_print();
}