#include "..\stdafx.h"
#include "windows.h"
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <map>
#include "graph.h"
#include "graphalgo.h"
#include "deadline.h"
#include "task_allocation.h"
#include "system.h"
#include "debug.h"
#include "schedule.h"
#include "table.h"
#include "branch_bound.h"
#include "main_extern.h"
#include "priority_assign.h"

Graph rand_grp;		//graph and system generated randomly
System rand_sys;
Graph GME_grp;		//graph and system from GME model
System GME_sys;	
Graph grp;			//graph and system used in algorithms - either from random or GME systems and models
System sys;
Graph meta_grp;
IBP_Policy policy;
bool allocdone;
Table<TaskType*, bool> dep_table;
CString lout;
CString mbuff;

using namespace std;

// Local functions

typedef struct
{
	TaskType* ptask;
	int priority;
} old_priority_data;

static list<old_priority_data> input_priorities;

static void print_statistics(allocation_stats&, System&);
static void print_proc_tasks(System&);
static void save_task_priority(list<old_priority_data>&, Graph&);
static void restore_task_priority(list<old_priority_data>&);
static void fill_task_rates ();
static bool check_task_rates ();

/*********************** HKL Testing *****************************/

static void test_HKL ();
static void algo_initialize();
static void map_tasks_proc();

void Do_Priority_Mapping ()
{
    SA_parameters sa;
	algo_initialize();

    fill_task_rates();
    if(!check_task_rates()) return;

    sa.initial_temperature=1000;
    sa.temp_const=0.995f;
    sa.equillibrium_jumps=100;
    sa.equillibrium_interval=100;
    sa.min_deadline_constant=2000;
    sa.thread_comm_constant=100;
    sa.distinct_level_constant=100;
    sa.stopping_jumps=500;
    sa.solution=false;

    map_tasks_proc();
    list<Processor*>::iterator ppi;
    
    for(ppi=sys.Proc_Begin();ppi!=sys.Proc_End();ppi++)
    {
        Processor* pp=*ppi;

        SA_priority_assign(pp,sa);
    }
    
    Print_Priority_Task_Map();
    return;
}
    
void Do_Period_Mapping ()
{
    map<TaskType*, TaskType*> task_map;
    list<TaskType*>::const_iterator ti,ti2;
    list<Processor*>::const_iterator pi;
    
    fill_task_rates();

    for(pi=sys.Proc_Begin();pi!=sys.Proc_End();pi++)
    {
        Processor *pproc=*pi;

        pproc->Clear_Tasks();
        pproc->Clear_Connections();
    }
    
    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()!=pt) continue;
        map<TaskType*, TaskType*>::const_iterator mi=task_map.find(pt);
        if(mi!=task_map.end()) continue;
        task_map[pt]=pt;
        for(ti2=grp.Node_Begin();ti2!=grp.Node_End();ti2++)
        {
            TaskType* pt2=*ti2;

            if(pt2->Property()->Get_Global_Root()!=pt2) continue; 
            if((pt2->Property()->Get_Period()==pt->Property()->Get_Period())&&
                    (pt2->Property()->Get_Alloc_Proc()==pt->Property()->Get_Alloc_Proc()))
            {
                task_map[pt2]=pt;
                pt2->Property()->Set_Priority(-1);
            }
        }
        
        Processor* pproc=pt->Property()->Get_Alloc_Proc();
        pproc->Add_Task(pt);
        pt->Property()->Set_Proc(pproc);
    }

    for(pi=sys.Proc_Begin();pi!=sys.Proc_End();pi++)
    {
        Processor *pproc=*pi;

        set_rma_priority(pproc);
    }

    map<TaskType*,TaskType*>::iterator mi;

    for(mi=task_map.begin();mi!=task_map.end();mi++)
    {
        TaskType* pt1=mi->first;
        TaskType* pt2=mi->second;

        int pr=pt2->Property()->Get_Priority();
        pt1->Property()->Set_Topo_Order();
        list<TaskType*>* topol=pt1->Property()->Topo();
        for(ti2=topol->begin();ti2!=topol->end();ti2++)
        {
            TaskType* pt3=*ti2;

            pt3->Property()->Set_Priority(pr);
        }
    }
    Print_Period_Task_Map(task_map);
}
    
void debug_print()
{
    mbuff=mbuff+lout;
}

void main_clear()
{
	input_priorities.clear();
}

void save_input_task_priorities()
{
	if(input_priorities.size()!=0) return;
	save_task_priority(input_priorities,grp);
	return;
}

void restore_input_task_priorities()
{
	if(input_priorities.size()==0) return;
	restore_task_priority(input_priorities);
	return;
}

static void print_proc_tasks(System& s)
{
    list<Processor*>::iterator ppi;
    list<TaskType*>::iterator pti;

    for(ppi=s.Proc_Begin();ppi!=s.Proc_End();ppi++)
    {
        Processor* pp=*ppi;

        lout.Format("\r\nProcessor %s\r\n", pp->Property()->Get_Name());
		debug_print();

        for(pti=pp->Tasks_Begin();pti!=pp->Tasks_End();pti++)
        {
            TaskType *pt=*pti;

            Print_Task(pt);
        }
    }
}

static void save_task_priority(list<old_priority_data>& olist, Graph& g)
{
	list<TaskType*>::iterator ti;

	for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
	{
		old_priority_data op;

		op.ptask=*ti;
		op.priority=op.ptask->Property()->Get_Priority();
		olist.push_front(op);
	}
}

static void restore_task_priority(list<old_priority_data>& plist)
{
	list<old_priority_data>::iterator oi;

	for(oi=plist.begin();oi!=plist.end();oi++)
	{
		old_priority_data od=*oi;

		TaskType* pt=od.ptask;
		pt->Property()->Set_Priority(od.priority);
	}
}

#define MAX_VALUE (float)99999999

static void fill_task_rates ()
{
    list<TaskType*>::iterator ti;
    int connected_count=1;
    
    /* First fill meta task rates */

    for(ti=meta_grp.Node_Begin();ti!=meta_grp.Node_End();ti++)
    {
        TaskType* pt=*ti;
        list<TaskType*> conn_list;
        float period=pt->Property()->Get_Period();

        if(period==0)
        {
            list<TaskType*>::iterator ti2;
            period=MAX_VALUE;

            get_pred_task_list(conn_list,pt);
            get_succ_task_list(conn_list,pt);
            conn_list.unique();

            for(ti2=conn_list.begin();ti2!=conn_list.end();ti2++)
            {
                TaskType* pt2=*ti2;

                if(pt2->Property()->Get_Period()<period)
                    period=pt2->Property()->Get_Period();
            }
            
            if((period==0)||(period==MAX_VALUE))
            {
                period=-(connected_count++);
            }

            pt->Property()->Set_Period(period);
        }
    }

    /* Fill task rates */

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType *pt=*ti;

        if(pt->Property()->Get_ID()==0) continue;
        if(pt->Get_Number_Of_Out_Connection()!=0) continue;
        
        list<TaskType*> pred_list;
        list<TaskType*>::iterator pi;
        float period=pt->Property()->Get_Period();
        float deadline=0;

		if(period==0)
		{
			period=pt->Property()->Get_Meta_Pointer()->Property()->Get_Period();
            deadline=pt->Property()->Get_Meta_Pointer()->Property()->Get_Deadline();
			pt->Property()->Get_Global_Property()->Set_Global_Period(period);
            pt->Property()->Get_Global_Property()->Set_Global_Deadline(deadline);
		}

        get_pred_task_list(pred_list,pt);
        pred_list.unique();
        for(pi=pred_list.begin();pi!=pred_list.end();pi++)
        {
            TaskType* pparent=*pi;

            if(pparent->Property()->Get_Period()==0)
                pparent->Property()->Set_Period(period);
        }
    }
    return;
}

static void test_HKL ()
{
    list<Processor*>::iterator pi;
    Processor* pproc=NULL;
    list<TaskType*>::iterator ti;

    for(pi=sys.Proc_Begin();pi!=sys.Proc_End();pi++)
    {
        pproc=*pi;
        break;
    }

    pproc->Clear_Tasks();

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* ptask=*ti;

        ptask->Property()->Set_Proc(pproc);
        pproc->Add_Task(ptask);
    }

    Schedule_Task(pproc);
    return;
}

static void algo_initialize()
{
    list<TaskType*>::iterator ti;

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()==pt)
        {
            pt->Property()->Set_Topo_Order();
            pt->Property()->Set_Global_Exe_Time();
        }
    }
}

static void map_tasks_proc()
{
    list<TaskType*>::iterator ti;
    list<Processor*>::iterator pi;

    for(pi=sys.Proc_Begin();pi!=sys.Proc_End();pi++)
    {
        Processor *pproc=*pi;

        pproc->Clear_Tasks();
        pproc->Clear_Connections();
    }

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;

        Processor* pproc=pt->Property()->Get_Alloc_Proc();
        pproc->Add_Task(pt);
        pt->Property()->Set_Proc(pproc);
    }
}

static bool check_task_rates ()
{
    list<TaskType*>::iterator ti;

    for(ti=grp.Node_Begin();ti!=grp.Node_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Period()<=0)
        {
            AfxMessageBox("Please enter period for "+pt->Property()->Get_Meta_Pointer()->Property()->Get_Name());
            return false;
        }
    }
    return true;
}
           

    

    
    
        
        
        

    
