#include "..\stdafx.h"
#include "windows.h"
#include <stdio.h>
#include <math.h>
#include "property.h"
#include "main_extern.h"
#include "graphalgo.h"

void TaskRTProperty::Debug()
{
    lout.Format("Ex %3.2f P %3.2f O %3.2f D %3.2f R %3.2f P %3d\r\n",exe_time, period, offset,
            deadline, response_time,priority);
    return;
}

void ConnectionRTProperty::Debug()
{
    lout.Format("S %3d P %3d\r\n",size, priority);
    return;
}
    
void TaskGlobalProperty::Set_Topo_Order ()
{
    Topo_Order.clear();
    Priority_Topo_Sort(Topo_Order,root,grp);
    return;
}

void TaskGlobalProperty::Canonize ()
{
    list<TaskType*>::reverse_iterator ti;
    int hp;
    
    ti=Topo_Order.rbegin();
    if(ti!=Topo_Order.rend()) hp=(*ti)->Property()->Get_Priority();
    for(ti++;ti!=Topo_Order.rend();ti++)
    {
        TaskType* pt=*ti;
        int op=pt->Property()->Get_Priority();
        
        if(op>hp) pt->Property()->Set_Priority(hp);
        else hp=op;
    }
}

int TaskGlobalProperty::Get_Min_Priority ()
{
    list<TaskType*>::iterator ti;
    int min_pr=100000000;

    for(ti=Topo_Order.begin();ti!=Topo_Order.end();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Priority()<min_pr) min_pr=pt->Property()->Get_Priority();
    }
    return min_pr;
}

int TaskGlobalProperty::Get_Max_Priority ()
{
    list<TaskType*>::iterator ti;
    int max_pr=-100000000;

    for(ti=Topo_Order.begin();ti!=Topo_Order.end();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Priority()>max_pr) max_pr=pt->Property()->Get_Priority();
    }
    return max_pr;
}

float TaskGlobalProperty::Get_Global_Response_Time ()
{
    TaskType* pt=Topo_Order.back();
    return pt->Property()->Get_Response_Time();
}

void TaskGlobalProperty::Set_Global_Exe_Time ()
{
    exe_time=0;
    list<TaskType*>::iterator ti;

    for(ti=Topo_Order.begin();ti!=Topo_Order.end();ti++)
    {
        TaskType* pt=*ti;

        exe_time+=pt->Property()->Get_Exec_Time();
    }
}

bool TaskProperty::Changed ()
{
    if(!pglobal) return false;
    
    list<TaskType*>& tl=pglobal->Topo_Order;
    list<TaskType*>::iterator ti;
    
    for(ti=tl.begin();ti!=tl.end();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Changed_MP()) return true;
    }

    if(pglobal->Changed_SP()) return true;
    else return false;
}

void TaskProperty::Restore ()
{
    if(!pglobal) return;
    
    list<TaskType*>& tl=pglobal->Topo_Order;
    list<TaskType*>::iterator ti;

    for(ti=tl.begin();ti!=tl.end();ti++)
    {
        TaskType* pt=*ti;

        pt->Property()->Restore_MP();
    }
    pglobal->Restore();
}

    
int TaskProperty::Get_Context_Switches()
{
    int csw=0;
    float resp_time=TaskRTProperty::Get_Response_Time();
    
    if(!palgo) return 0;
    
    set<TaskType*> pmp=palgo->MP;
    set<TaskType*>::iterator ti;
    for(ti=pmp.begin();ti!=pmp.end();ti++)
    {
        TaskType* pt1=*ti;

        csw+=ceil(resp_time/pt1->Property()->Get_Period());
    }
    
    return csw;
}

TaskType* TaskProperty::Get_Predecessor()
{
    if(!pglobal) return NULL;
    
    list<TaskType*>& tl=pglobal->Topo_Order;
    list<TaskType*>::iterator ti;
    TaskType* parent=NULL;
	TaskType* myp=TaskIDProperty::Task();

    for(ti=tl.begin();ti!=tl.end();ti++)
    {
       TaskType* pt=*ti;

       if(pt==myp) break;
       else parent=pt;
    }
    return parent;
}

int TaskProperty::Get_First_Priority()
{
    int myp= TaskRTProperty::Get_Priority();
    
    if(!pglobal) return myp;
    
    list<TaskType*>& tl=pglobal->Topo_Order;
    list<TaskType*>::iterator ti=tl.begin();

    if(ti!=tl.end()) return (*ti)->Property()->Get_Priority();
    else return myp;
}

int TaskProperty::Get_Last_Priority()
{
    int myp= TaskRTProperty::Get_Priority();
    
    if(!pglobal) return myp;
    
    list<TaskType*>& tl=pglobal->Topo_Order;
    list<TaskType*>::reverse_iterator ti=tl.rbegin();

    if(ti!=tl.rend()) return (*ti)->Property()->Get_Priority();
    else return myp;
}

void TaskProperty::Restore_User_Priorities()
{
    list<TaskType*>::iterator ti;
    list<TaskType*>& topol=pglobal->Topo_Order;
    
    if(!pglobal) return;
    
    for(ti=topol.begin();ti!=topol.end();ti++)
    {
        TaskType* pt1=*ti;

        int oldp=pt1->Property()->Get_User_Priority();
        pt1->Property()->Set_Priority(oldp);
    }
    return;
}

bool ProcessorProperty::Allocate_Space(double utilization)
{
	if(utilization > available_space)
		return false;

	available_space -= utilization;
	return true;
}

bool ProcessorProperty::Temp_Allocate_Space(double utilization)
{
	if(utilization > temp_available_space)
		return false;

	temp_available_space -= utilization;
	return true;
}

void ProcessorProperty::Set_Bound(double b)
{
	bound = b;
	available_space = b;
}
