#ifndef SYSTEM_H_0
#define SYSTEM_H_0

#include "processor_network.h"

class System
{
    private:
        list<Processor *> processor_list;
        list<Network *> network_list;
        list<OS *> oses;
		CString description;
    public:
        System(){};
        inline list<Processor *>::iterator Proc_Begin(){return processor_list.begin();};
        inline list<Processor *>::iterator Proc_End(){return processor_list.end();};
        inline list<Network *>::iterator Net_Begin(){return network_list.begin();};
        inline list<Network *>::iterator Net_End(){return network_list.end();};
		inline list<OS *>::iterator OS_Begin(){return oses.begin();};
        inline list<OS *>::iterator OS_End(){return oses.end();};
        inline int Get_Number_Of_Procs () {return processor_list.size();};
        inline int Get_Number_Of_Nets () {return network_list.size();};
        inline int Get_Number_Of_OSes () {return oses.size();};
        inline Network * Get_Network() {return network_list.front();};
        inline OS* Get_OS(){return oses.front();};
        inline void Add_Proc(Processor *p){processor_list.push_front(p);};
        inline void Add_Net(Network *n){network_list.push_front(n);};
        inline void Add_OS(OS *o){oses.push_front(o);}; 
        inline ProcessorProperty* Get_Generic_Processor_Property()
        {
            if(!processor_list.empty())
                return processor_list.front()->Property();
            else return NULL;
        }
        void Order_Procs_By_CAN_ID();
		void Order_Procs_By_Name();		//Jeff Added
        Processor* Get_Proc_By_Name(CString&);
	    Network* Get_Network_By_Name(CString&);
        OS* Get_OS_By_Name(CString&);                
        void Clean();

		//Jeff Added
		inline CString GetDescription()				{return description;}
		inline void SetDescription(CString desc)	{description = desc;}
};


#endif
