#ifndef TASK_CONNECTION_H_0
#define TASK_CONNECTION_H_0

#include <list>
#include "property.h"

using namespace std;

class TaskType;
class ConnectionType
{
    private:
        ConnectionProperty *property;
        TaskType *to;
        TaskType *from;
    public:
        ConnectionType(){property=NULL;from=NULL;to=NULL;};
        ~ConnectionType(){if(property) delete property;}
        inline TaskType * Get_From () {return from;};
        inline TaskType * Get_To () {return to;};
        inline ConnectionProperty * Property () {return property;};
        inline void Set_From (TaskType *t) {from=t;};
        inline void Set_To (TaskType *t) {to=t;};
        inline void Set_Property (ConnectionProperty *p) {property=p;};
};

/* There are 2 different types of connections for tasks: Data connections and 
 * event connections.  Data connections deal with data being passed between 
 * different tasks and are indicated by the word 'Data' in the function names.
 * Event connections are simply signals between tasks and are the 'default'
 * function names.
 */
class TaskType
{
    private:
        list<ConnectionType *> inedges;
        list<ConnectionType *> outedges;
		list<ConnectionType *> inDataEdges;	 //Jeff added
		list<ConnectionType *> outDataEdges; //Jeff added
        TaskProperty *property;
    public:
        TaskType(){property=NULL;};
        ~TaskType(){if(property) delete property;}
        inline void Add_In_Connection (ConnectionType *e) {inedges.push_front(e);};
        inline void Add_Out_Connection (ConnectionType *e) {outedges.push_front(e);};
        inline int Get_Number_Of_In_Connection () { return inedges.size();};
        inline int Get_Number_Of_Out_Connection () { return outedges.size();};
        inline TaskProperty * Property () {return property;};
        inline void Set_Property (TaskProperty *p) {property=p;};
        inline list<ConnectionType *>::iterator Get_Out_Edges_Begin(){return outedges.begin();};
        inline list<ConnectionType *>::iterator Get_Out_Edges_End(){return outedges.end();};
        inline list<ConnectionType *>::iterator Get_In_Edges_Begin(){return inedges.begin();};
        inline list<ConnectionType *>::iterator Get_In_Edges_End(){return inedges.end();};
        ConnectionType* Get_Connection_To (TaskType*);
        ConnectionType* Get_Connection_From (TaskType*);

		//Jeff added functionality for data connections
		inline void Add_In_Data_Connection (ConnectionType *e) {inDataEdges.push_front(e);}
        inline void Add_Out_Data_Connection (ConnectionType *e) {outDataEdges.push_front(e);}
		inline int Get_Number_Of_In_Data_Connection () { return inDataEdges.size();}
        inline int Get_Number_Of_Out_Data_Connection () { return outDataEdges.size();}
		inline list<ConnectionType *>::iterator Get_Out_Data_Edges_Begin(){return outDataEdges.begin();};
        inline list<ConnectionType *>::iterator Get_Out_Data_Edges_End(){return outDataEdges.end();};
        inline list<ConnectionType *>::iterator Get_In_Data_Edges_Begin(){return inDataEdges.begin();};
        inline list<ConnectionType *>::iterator Get_In_Data_Edges_End(){return inDataEdges.end();};
		ConnectionType* Get_Data_Connection_To (TaskType*);
        ConnectionType* Get_Data_Connection_From (TaskType*);
};

#endif
