// component_dialog.cpp : implementation file
//

#include <math.h>
#include <stdio.h>

#include "..\stdafx.h"
#include "..\component.h"
#include "component_dialog.h"
#include "..\linux_port\main_extern.h"
#include "..\linux_port\first_fit.h"
#include "..\linux_port\load_balance.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// component_dialog dialog


component_dialog::component_dialog(CWnd* pParent /*=NULL*/)
	: CDialog(component_dialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(component_dialog)
	//}}AFX_DATA_INIT
}


void component_dialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(component_dialog)
	DDX_Control(pDX, IDC_EDIT2, m_edit);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(component_dialog, CDialog)
	//{{AFX_MSG_MAP(component_dialog)
	ON_BN_CLICKED(IDC_RADIO1, OnSubsystemMap)
	ON_BN_CLICKED(IDC_RADIO2, OnPriorityMap)
	ON_BN_CLICKED(IDC_BUTTON1, OnDoMap)
	ON_BN_CLICKED(IDC_RADIO3, OnFirstFitMap)
	ON_BN_CLICKED(IDC_RADIO4, OnLoadBalanceMap)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// component_dialog message handlers

void component_dialog::OnSubsystemMap() 
{
	policy.mp=PERIOD_MAPPING;
}

void component_dialog::OnPriorityMap() 
{
	policy.mp=PRIORITY_MAPPING;
}

void component_dialog::OnFirstFitMap() 
{
	policy.mp=FIRSTFIT_MAPPING;
}

void component_dialog::OnLoadBalanceMap() 
{
	policy.mp=LOAD_BALANCE_MAPPING;
}

void component_dialog::OnDoMap() 
{
	mbuff.Empty();

	double r;
	
    switch(policy.mp)
    {
        case PRIORITY_MAPPING: 
            Do_Priority_Mapping();
            break;
        case PERIOD_MAPPING:
            Do_Period_Mapping();
            break;
		case FIRSTFIT_MAPPING:
			r = Do_FirstFit_Mapping();	//returns 0 if everything is ok, if not it returns
										//the utilization that could not be put on a processor

			if(r)						//error
				show_Mapping_Error(r);
			break;
		case LOAD_BALANCE_MAPPING:
			r = Do_LoadBalance_Mapping();	//returns 0 if everything is ok, if not it returns
											//the utilization that could not be put on a processor

			if(r)						//error
				show_Mapping_Error(r);
            break;
    }
    
	m_edit.SetWindowText(mbuff);
	return;
}

void show_Mapping_Error(double extra_utilization)
{
	string msg;

	char extra_procs_10[10];	//for bound of 1.0
	char extra_procs_08[10];	//for bound of 0.8
	char extra_procs_06[10];	//for bound of 0.6
	char extra_procs_04[10];	//for bound of 0.4

	//multiply extra_utilization by the inverse of the bound
	sprintf(extra_procs_10, "%i", (int)ceil(extra_utilization));
	sprintf(extra_procs_08, "%i", (int)ceil(extra_utilization * 1.25));
	sprintf(extra_procs_06, "%i", (int)ceil(extra_utilization * 1.67));
	sprintf(extra_procs_04, "%i", (int)ceil(extra_utilization * 2.5));

	msg = string("There are not enough processors available to process all of the tasks.") +
		"  Please adjust the bounds of the existing CPU's or add more processors.\n\n" + 
		"For a bound of 1.0:\t " + extra_procs_10 + " extra processor(s) are needed\n" +
		"For a bound of 0.8:\t " + extra_procs_08 + " extra processor(s) are needed\n" +
		"For a bound of 0.6:\t " + extra_procs_06 + " extra processor(s) are needed\n" +
		"For a bound of 0.4:\t " + extra_procs_04 + " extra processor(s) are needed\n";

	AfxMessageBox(msg.c_str(), MB_OK | MB_ICONSTOP);
}
