#include "first_fit.h"

//global includes
#include <set>

//local includes
#include "graph.h"
#include "system.h"
#include "debug.h"

using namespace std; 

//Externals
extern Graph grp;
extern System sys;

double Do_FirstFit_Mapping()
{
	multiset<TaskType*, task_greater_than>		 tasks;
	multiset<Processor*, processor_greater_than> processors;
	double										 non_allocated_utility = 0;

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		(*proc_it)->Clear_Tasks();							//clear all tasks on the processor
		(*proc_it)->Property()->Reset_Available_Space();	//give the processor its entire amount of space available
	}

	//automatically sorts in order of the utilization
	for(list<TaskType*>::iterator grp_it = grp.Node_Begin(); grp_it != grp.Node_End(); grp_it++)
		tasks.insert(*grp_it);

	//automatically sorts in order of bounds
	for(proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
		processors.insert(*proc_it);

	//map each task to a processor
	for(multiset<TaskType*, task_greater_than>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
	{
		double utilization = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();;

		for(multiset<Processor*, processor_greater_than>::iterator proc_it = processors.begin(); proc_it != processors.end(); ++proc_it)
		{
			//is there room for this task on this processor
			if((*proc_it)->Property()->Allocate_Space(utilization))
			{
				(*proc_it)->Add_Task(*task_it); 
				break;
			}
		}

		//could not find a processor to map the task to
		if(proc_it == processors.end())
			non_allocated_utility += utilization;
	}

	//were there any tasks that could not be mapped to a processor (error)
	if(non_allocated_utility)
		return non_allocated_utility;

	//print out the results
	print_FirstFit_Mapping();

	return 0;	//no errors
}


//Function object definitions
bool task_greater_than::operator () (TaskType* t1, TaskType* t2)
{
	double u1 = t1->Property()->Get_Exec_Time() / t1->Property()->Get_Period();
	double u2 = t2->Property()->Get_Exec_Time() / t2->Property()->Get_Period();

	return u1 > u2;
}

bool processor_greater_than::operator () (Processor* p1, Processor* p2)
{
	return p1->Property()->Get_Bound() > p2->Property()->Get_Bound();
}