#ifndef GRAPH_H_0
#define GRAPH_H_0

/*
* A template graph data structure supporting graph operation, mainly traversal.
* No exception handling mechanism is provided. 
*/

#include <list>
#include "property.h"
#include "task_connection.h"

template<class N, class E> class GenericGraph
{
    private:
        list<N *> nodes;
        list<E *> edges;
    public:
        GenericGraph(){};
        inline int Get_Number_Of_Nodes(){return nodes.size();};
        inline int Get_Number_Of_Edges(){return edges.size();};
        inline void Add_Node (N *n){nodes.push_front(n);return;};
        inline void Add_Edge (E *e){edges.push_front(e);return;};
        list<N *> Get_Root_Nodes();
        list<N *> Topology_Sort ();
        void BFS_Traversal (void (*)(N *, N *,E *));
        void DFS_Traversal (void (*)(N *, N *,E *));
        GenericGraph *Clone ();
        inline list<N *>::iterator Node_Begin(){return nodes.begin();};
        inline list<N *>::iterator Node_End(){return nodes.end();};
        inline list<E *>::iterator Edge_Begin(){return edges.begin();};
        inline list<E *>::iterator Edge_End(){return edges.end();};
        inline void Clear() {nodes.clear();edges.clear();};
        N* Get_Node_By_Name (CString&);
        void Clean();
};

typedef GenericGraph<TaskType,ConnectionType> Graph;

#endif
