#ifndef GRAPHALGO_H_0
#define GRAPHALGO_H_0

#include "graph.h"
#include "table.h"

#define NIL	            -1

/* 
* The TaskProperty and ConnectionProperty should support function
* Get_Exec_Time() and Get_Trans_Time(). The Graph, TaskType, ConnectionType
* are as in graph.h.
*/

typedef struct
{
	float distance; /* Max distance from source */
    int pathlength; /* Path length from source */
	ConnectionType *pinedge;
    TaskType *ptask;
} sssp_entry;

extern list<sssp_entry>* SSSP_Dijkstra (Graph&, TaskType&);
extern list<sssp_entry>* SSSP_Bellman_Ford (Graph&, TaskType&);

/*
*  These take nodes whose Get_Number_Of_In_Connection() returns zero
* as sources.
*/

extern list<sssp_entry>* SSSP_Dijkstra (Graph&);
extern list<sssp_entry>* SSSP_Bellman_Ford (Graph&);
extern TaskType* get_parent_task(TaskType *, list<sssp_entry>*);

extern list<TaskType *>* Topo_Sort (Graph&);
extern void Topo_Sort (list<TaskType*>&,TaskType*,Graph&);
extern void Priority_Topo_Sort (list<TaskType*>&,TaskType*,Graph&);
extern void Topo_Sort (list<TaskType*>&,TaskType*,list<TaskType*>&);
extern void BFS_Traversal(Graph&,TaskType*,
        void (*)(TaskType*,TaskType*,ConnectionType*));

extern void get_pred_task_list(list<TaskType*>&,TaskType*);
extern void get_succ_task_list(list<TaskType*>&,TaskType*);
extern void get_sub_graph(Graph&, list<TaskType*>&);
extern void set_dependency_table(Table<TaskType*,bool>&, Graph&);

#endif

