#ifndef LOAD_BALANCE_H
#define LOAD_BALANCE_H

/* Load Balance Algorithm
 * ----------------------
 *
 * Given a set of processors and a set of tasks to be completed on these processors 
 * divide the work as evenly as possibly between all of the processors.  This is done
 * by first sorting all of the tasks by there utilization.  Utilization is found by
 * dividing the execution time of the task by the period of the task.  The processors
 * are also sorted by the total amount of utilization all of the tasks that have been
 * assigned to it have.  The task with the highest utilization is then assigned to the
 * processor with the lowest utilization that has enough space to take on the task.  After
 * this the processors are resorted and the rest of the tasks are processed the same
 * way until there are no more.  If a task cannot be mapped to any processor then this
 * is an error and the bounds of the existing processors need to be adjusted or more
 * processors need to be added.
 */

class Processor;

//map the tasks acording to the load balance algorithm
double Do_LoadBalance_Mapping();

//Function object to determine whether or not the total utilization of all the tasks on
//processor p1 is less then the the total utilization on processor p2 (used for sorting) 
class processor_less_than
{
public:
	bool operator () (Processor* p1, Processor* p2);
};

#endif