#include "..\stdafx.h"
#include "windows.h"
#include <stdio.h>
#include <math.h>
#include <algorithm>
#include "schedule.h"
#include "debug.h"
#include "graphalgo.h"
#include "system.h"
#include "main_extern.h"

#undef max
#define max(x,y) (((x)>(y))?(x):(y))
#undef min
#define min(x,y) (((x)<(y))?(x):(y))

static void initialize_task_list_dependency (list<TaskType*>&);
static void SPPS (list<TaskType *>&, Processor *);
static float mod (float, float);
static void initialize_messages(list<ConnectionType*>&);
static void find_message_response_time(ConnectionType*,list<ConnectionType*>&,
        Network*);
static void find_response_time_execution(TaskType*,list<TaskType*>&,Processor*);
static void set_children_offset(TaskType*,list<TaskType*>&);
static void find_response_time_execution(TaskType*,list<TaskType*>&, 
        Processor*,Table<TaskType*,float>&);
static void order_messages_period (list<ConnectionType*>&, list<ConnectionType*>&);

/******************************* HKL *************************************/

static void populate_mp_sp (Processor *);
static void populate_mp_sp (Processor *,TaskType*);
static bool is_mp(TaskType*,TaskType*);
static bool is_sp(TaskType*,TaskType*);
static bool is_lp(TaskType* pt1, TaskType* pt2);
static void find_response_time (TaskType*);
static float find_length_busy_period (TaskType*);
static float get_blocking_time (TaskType*);
static float sum_h_segments (TaskType*,TaskType*);
static float sum_h_inner_segments(TaskType*,TaskType*);
static float sum_h_initial_segments(TaskType*,TaskType*);
static float sum_h_final_segments(TaskType*,TaskType*);
static void get_SP (set<TaskType*>&, TaskType*);
static float find_response_time_sub_task(TaskType*,float);
static float find_response_time_first_task (TaskType*,int);
static float find_length_busy_period (TaskType*);
static void order_tasks_list (list<TaskType*>&,list<TaskType*>&);
static bool check_utilization (Processor *);

/********************* TESTING
 
void Schedule_Task (Processor *pp)
{
    list<TaskType*> tl;
    list<TaskType*>::iterator si;

    for(si=pp->Tasks_Begin();si!=pp->Tasks_End();si++)
    {
        TaskType *ptt=*si;

        tl.push_front(ptt);
    }

    SPPS(tl,pp);
}

void Schedule_Task (Processor *pproc)
{
    list<TaskType*>::iterator ti;

    if(!check_utilization(pproc)) return;
    populate_mp_sp(pproc);
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if((pt->Property()->Get_Global_Root()==pt)
                && pt->Property()->Changed())
				find_response_time(pt);
    }
    return;
}

                       TESTING ***********************/

/******************** HKL ************************/

void Schedule_Task (Processor *pproc)
{
    list<TaskType*>::iterator ti;

    if(!check_utilization(pproc)) return;
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if((pt->Property()->Get_Global_Root()==pt)
                && pt->Property()->Changed())
        {
                populate_mp_sp(pproc,pt);
				find_response_time(pt);
                // pt->Property()->Restore_User_Priorities();
        }
    }
    return;
}

void Schedule_CAN_Messages (System& sys)
{
    list<ConnectionType*> sys_msgs, proc_msgs, ordered_proc_msgs;
    list<Processor*>::iterator ip;
    list<ConnectionType*>::iterator ic;
    
    sys.Order_Procs_By_CAN_ID();
    Network *net=sys.Get_Network();

    for(ip=sys.Proc_Begin();ip!=sys.Proc_End();ip++)
    {
        Processor* pproc=*ip;

        proc_msgs.clear();
        ordered_proc_msgs.clear();
        for(ic=pproc->Connection_Begin();ic!=pproc->Connection_End();ic++)
        {
            proc_msgs.push_front(*ic);
        }

        order_messages_period(ordered_proc_msgs, proc_msgs);

        for(ic=ordered_proc_msgs.begin();ic!=ordered_proc_msgs.end();ic++)
        {
            sys_msgs.push_back(*ic);
        }
    }

    initialize_messages(sys_msgs);

    for(ic=sys_msgs.begin();ic!=sys_msgs.end();ic++)
    {
        ConnectionType *pc=*ic;

        find_message_response_time(pc,sys_msgs,net);
    }
}

void Split_Interference(Processor *pproc, Table<TaskType*,float>& proc_table)
{
    list<TaskType*>::iterator it;
    list<TaskType*> tl,task_list;
    int num_tasks=pproc->Get_Number_Of_Tasks();
    
    proc_table.Set_Number_Of_Elements(num_tasks);
    for(it=pproc->Tasks_Begin();it!=pproc->Tasks_End();it++)
    {
        TaskType* pt=*it;

        tl.push_front(pt);
    }
    
    initialize_task_list_dependency(tl);

    order_tasks_dependency(task_list,tl);
    
    for(it=task_list.begin();it!=task_list.end();it++)
    {
        TaskType* pt=*it;
        
        proc_table.Add_Element(pt);
    }
    proc_table.Initialize(0);

    for(it=task_list.begin();it!=task_list.end();it++)
    {
        TaskType* pt=*it;
        
        find_response_time_execution(pt,task_list,pproc,proc_table);
        set_children_offset(pt,task_list);
    }
}

void order_tasks_dependency (list<TaskType*>& task_list, list<TaskType*>&tl)
{
    list<TaskType*>::iterator otli;
    list<TaskType*>::iterator otask_listi;
    bool inserted;
    
    for(otli=tl.begin();otli!=tl.end();otli++)
    {
        TaskType *pt=*otli;
        inserted=false;
        
        for(otask_listi=task_list.begin();otask_listi!=task_list.end();otask_listi++)
        {
            TaskType *pt2=*otask_listi;
            float rel_deadline1, rel_deadline2;

            switch(policy.sp)
            {
                case PRIORITY_POLICY: // Deliberately inverted.
                    rel_deadline1=pt2->Property()->Get_Priority();
                    rel_deadline2=pt->Property()->Get_Priority();
                    break;
                case ADVANCED_RMA:
                default:
                    rel_deadline1=pt->Property()->Get_Period();
                    rel_deadline2=pt2->Property()->Get_Period();
                    break;
            }

            
            if(rel_deadline1<rel_deadline2)
            {
                task_list.insert(otask_listi,pt);
                inserted=true;
                break;
            }
            if(rel_deadline1==rel_deadline2)
            {
                bool b;
                if(dep_table.Get_Value(pt,pt2,b))
                    if(b)
                    {
                        task_list.insert(otask_listi,pt);
                        inserted=true;
                        break;
                    }
            }
        }
        if(!inserted) task_list.push_back(pt);
    }
}

float Get_Task_Utilization(Processor *pp)
{
    list<TaskType*>::iterator ti;
    float util=0;
    
    for(ti=pp->Tasks_Begin();ti!=pp->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        util+=(pt->Property()->Get_Exec_Time()/pt->Property()->Get_Period());
    }
    return util;
}

float Get_Scheduling_Utilization (Processor *pp)
{
    list<TaskType*>::iterator ti;
    float cost=pp->Get_OS()->Property()->Get_Scheduling_Overhead();
    float util=0;
    
    for(ti=pp->Tasks_Begin();ti!=pp->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        util+=(cost/pt->Property()->Get_Period());
    }
    return util;
}

float Get_Context_Switch_Utilization (Processor *pp)
{
    list<TaskType*>::iterator ti;
    float util=0;
    float cost=pp->Get_OS()->Property()->Get_Context_Switch_Overhead();

    for(ti=pp->Tasks_Begin();ti!=pp->Tasks_End();ti++)
    {
        TaskType* pt=*ti;
        int task_csw=0;
        set<TaskType*>* pmp=pt->Property()->MP();
        set<TaskType*>::iterator ti;
        float resp_time=pt->Property()->Get_Response_Time();

        for(ti=pmp->begin();ti!=pmp->end();ti++)
        {
            TaskType* pt1=*ti;

            task_csw+=ceil(resp_time/pt1->Property()->Get_Period());
        }
        
        util+=(task_csw*cost/pt->Property()->Get_Period());
    }
    return util;
}

float Get_Timer_Utilization (Processor *pp)
{
    OS* pos=pp->Get_OS();
    
    return (pos->Property()->Get_Timer_Resolution()/1000)*
        pos->Property()->Get_Timer_Overhead();
}

float Get_Network_Utilization (System& sys)
{
    list<ConnectionType*> sys_msgs;
    list<Processor*>::iterator ip;
    list<ConnectionType*>::iterator ic;
    
    for(ip=sys.Proc_Begin();ip!=sys.Proc_End();ip++)
    {
        Processor* pproc=*ip;

        for(ic=pproc->Connection_Begin();ic!=pproc->Connection_End();ic++)
        {
            sys_msgs.push_front(*ic);
        }
    }

    float util=0;
    
    for(ic=sys_msgs.begin();ic!=sys_msgs.end();ic++)
    {
        ConnectionType *pc=*ic;

        util+=(pc->Property()->Get_Trans_Time()/pc->Get_To()->Property()->Get_Period());
    }
    return util;
}

void set_rma_priority (Processor *pproc)
{
    list<TaskType*> task_list, tl;
    list<TaskType*>::iterator ti;
    
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
        tl.push_front(*ti);

    order_tasks_dependency(task_list,tl);
    int max_p=task_list.size();
    for(ti=task_list.begin();ti!=task_list.end();ti++)
    {
        TaskType* pt=*ti;

        pt->Property()->Set_Priority(max_p);
        max_p--;
    }
    return;
}

void set_rma_priority (System& s)
{
	list<Processor*>::iterator pi;

	for(pi=s.Proc_Begin();pi!=s.Proc_End();pi++)
	{
		Processor *pproc=*pi;

		set_rma_priority(pproc);
	}
}

void set_rma_priority (Graph& g)
{
	list<TaskType*>::iterator ti;
	list<TaskType*> task_list, tl;

	for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
		tl.push_front(*ti);

	order_tasks_dependency(task_list,tl);
    int max_p=task_list.size();
    for(ti=task_list.begin();ti!=task_list.end();ti++)
    {
        TaskType* pt=*ti;

        pt->Property()->Set_Priority(max_p);
        max_p--;
    }
    return;
}

static void initialize_task_list_dependency (list<TaskType*>& tl)
{
    list<TaskType*>::iterator ti;

    for(ti=tl.begin();ti!=tl.end();ti++)
    {
        TaskType* pt=*ti;

        pt->Property()->Set_Response_Time(MAX);
        //pt->Property()->Set_Offset(0);
        pt->Property()->Reset_Context_Switches();
    }
    return;
}

static void SPPS (list<TaskType *>& tl, Processor *p)
{
    list<TaskType*> task_list;
    list<TaskType*>::iterator si;
    
    initialize_task_list_dependency(tl);
    order_tasks_dependency(task_list,tl);

    for(si=task_list.begin();si!=task_list.end();si++)
    {
        TaskType* pt=*si;
        
        find_response_time_execution(pt,task_list,p);
        //set_children_offset(pt,task_list);
    }
}

static float mod (float x, float y)
{
    return x-floor(x/y)*y;
}

static void find_message_response_time(ConnectionType* pt, 
        list<ConnectionType*>& message_list, Network *net)
{
    float rppc, rppn;
    float interference_duration;
    float block_time=0;
    
    list<ConnectionType*>::iterator fi;
    
    fi=find(message_list.begin(),message_list.end(),pt);
    if(fi==message_list.end()) return;
    fi++;
    for(;fi!=message_list.end();fi++)
    {
        ConnectionType *pc=*fi;
        
        if(pc->Property()->Get_Trans_Time()>block_time)
            block_time=pc->Property()->Get_Trans_Time();
    }
    
    rppn=0;
    rppc=NIL;
    while((rppc!=rppn)&&(rppn < 3*(pt->Get_To()->Property()->Get_Period())))
    {
        float sum=0;
        rppc=rppn;

        for(fi=message_list.begin();fi!=message_list.end();fi++)
        {
            ConnectionType* pht=*fi;
            float intht;
            
            if(pht==pt) break;
            
            interference_duration=rppc;
            intht=ceil(interference_duration/
                    pht->Get_To()->Property()->Get_Period())*
                pht->Property()->Get_Trans_Time();
            sum+=intht;
        }
        rppn=sum+block_time;
    }
    rppn=rppn+pt->Property()->Get_Trans_Time();
    pt->Property()->Set_Response_Time(rppn);
    return;
}

static void initialize_messages(list<ConnectionType*>& msg_list)
{
    list<ConnectionType*>::iterator im;

    for(im=msg_list.begin();im!=msg_list.end();im++)
    {
        ConnectionType* pc=*im;

        pc->Property()->Set_Response_Time(MAX);
    }
    return;
}

static void find_response_time_execution(TaskType* pt, list<TaskType*>& task_list, 
        Processor *p)
{
    float rppc, rppn;
    float interference_duration;
    float jitter=0; // pt->Property()->Get_Offset();
    OS* pos=p->Get_OS();
    int csw;
    
    list<TaskType*>::iterator fi;
   
    if(pt->Property()->Get_Exec_Time()==0)
    {
        pt->Property()->Set_Response_Time(0);
        return;
    }

    rppn=0;
    rppc=NIL;
    while((rppc!=rppn)&&(rppn < (3*pt->Property()->Get_Period())))
    {
        float sum=0;
        csw=0;
        rppc=rppn;

        for(fi=task_list.begin();fi!=task_list.end();fi++)
        {
            TaskType* pht=*fi;
            float intht;
            float decrease=0;

			if(pht->Property()->Get_ID()==0) continue;
            
            if(pht==pt) break;
            if(pht->Property()->Get_Response_Time()<=
                    mod(jitter,pht->Property()->Get_Period()))
            {
                if(jitter>0)
                    decrease=pht->Property()->Get_Period()-mod(jitter,
                            pht->Property()->Get_Period());
            }

            interference_duration=max(0, rppc-decrease);
//            interference_duration=rppc;
            intht=ceil(interference_duration/pht->Property()->Get_Period())*
                (pht->Property()->Get_Exec_Time()+
                 2*pos->Property()->Get_Context_Switch_Overhead()+
                 pos->Property()->Get_Scheduling_Overhead());
            float timer_inter=ceil(interference_duration/
                    (1000/pos->Property()->Get_Timer_Resolution()))*
                pos->Property()->Get_Timer_Overhead();
            sum+=(intht+timer_inter);
            csw+=ceil(interference_duration/pht->Property()->Get_Period());
        }
        rppn=sum+pt->Property()->Get_Exec_Time();
    }
    pt->Property()->Set_Response_Time(rppn+jitter);
    pt->Property()->Set_Context_Switches(csw);
    return;
}

static void set_children_offset(TaskType* pt, list<TaskType*>& task_list)
{
    float resp_time=pt->Property()->Get_Response_Time();
    list<TaskType*>::iterator it;
    list<ConnectionType*>::iterator ic;

    for(ic=pt->Get_Out_Edges_Begin();ic!=pt->Get_Out_Edges_End();ic++)
    {
        ConnectionType* pc=*ic;

        if(!(pc->Property()->Is_Control_Flow())) continue;
        TaskType* pt2=pc->Get_To();
        it=find(task_list.begin(),task_list.end(),pt2);
        if(it!=task_list.end())
        {
            pt2->Property()->Set_Offset(resp_time);
        }
    }
}

static void find_response_time_execution(TaskType* pt, list<TaskType*>& task_list, 
        Processor *p, Table<TaskType*,float>& proc_table)
{
    float rppc, rppn;
    float interference_duration;
    float jitter=0; // pt->Property()->Get_Offset();
    OS* pos=p->Get_OS();
    int csw;
    
    list<TaskType*>::iterator fi;
   
    if(pt->Property()->Get_Exec_Time()==0)
    {
        pt->Property()->Set_Response_Time(0);
        return;
    }

    rppn=0;
    rppc=NIL;
    while((rppc!=rppn)&&(rppn < (3*pt->Property()->Get_Period())))
    {
        float sum=0;
        csw=0;
        rppc=rppn;

        for(fi=task_list.begin();fi!=task_list.end();fi++)
        {
            TaskType* pht=*fi;
            float intht;
            float decrease=0;
            
			if(pht->Property()->Get_ID()==0) continue;
            if(pht==pt) break;
            if(pht->Property()->Get_Response_Time()<=
                    mod(jitter,pht->Property()->Get_Period()))
            {
                if(jitter>0)
                    decrease=pht->Property()->Get_Period()-mod(jitter,
                            pht->Property()->Get_Period());
            }

            interference_duration=max(0, rppc-decrease);
//            interference_duration=rppc;
            intht=ceil(interference_duration/pht->Property()->Get_Period())*
                (pht->Property()->Get_Exec_Time()+
                 2*pos->Property()->Get_Context_Switch_Overhead()+
                 pos->Property()->Get_Scheduling_Overhead());
            float timer_inter=ceil(interference_duration/
                    (1000/pos->Property()->Get_Timer_Resolution()))*
                pos->Property()->Get_Timer_Overhead();
            sum+=(intht+timer_inter);
            csw+=ceil(interference_duration/pht->Property()->Get_Period());
            proc_table.Set_Value(pt,pht,intht);
        }
        rppn=sum+pt->Property()->Get_Exec_Time();
    }
    pt->Property()->Set_Response_Time(rppn+jitter);
    pt->Property()->Set_Context_Switches(csw);
    proc_table.Set_Value(pt,pt,rppn+jitter);
    return;
}

static void order_messages_period (list<ConnectionType*>& msg_list, 
        list<ConnectionType*>&tl)
{
    list<ConnectionType*>::iterator otli;
    list<ConnectionType*>::iterator omsg_listi;
    bool inserted;
    
    for(otli=tl.begin();otli!=tl.end();otli++)
    {
        ConnectionType *pt=*otli;
        inserted=false;
        
        for(omsg_listi=msg_list.begin();omsg_listi!=msg_list.end();omsg_listi++)
        {
            ConnectionType *pt2=*omsg_listi;
            float rel_deadline1, rel_deadline2;

			rel_deadline1=rel_deadline2=0;

			switch (policy.sp)
			{
                case PRIORITY_POLICY:// Deliberately inverted to obtain correct results. Why? think!!!
                    rel_deadline1=pt2->Get_To()->Property()->Get_Priority();
    				rel_deadline2=pt->Get_To()->Property()->Get_Priority();
                    break;
                case ADVANCED_RMA:
                default:
                    rel_deadline1=pt->Get_To()->Property()->Get_Period();
                    rel_deadline2=pt2->Get_To()->Property()->Get_Period();
                    break;
			}
            
            if(rel_deadline1<rel_deadline2)
            {
                msg_list.insert(omsg_listi,pt);
                inserted=true;
                break;
            }
        }
        if(!inserted) msg_list.push_back(pt);
    }
}

/********************************* HKL *****************************************/



void Restore_MP_SP (Processor *pproc)
{
    list<TaskType*>::iterator ti;

    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        if(pt->Property()->Get_Global_Root()==pt) pt->Property()->Restore();
    }
    return;
}

static bool check_utilization (Processor *pproc)
{
    list<TaskType*>::iterator ti;
    float u=0;
    
    /******************* OverHeads *******************/
    OS* pos=pproc->Get_OS();
    float csoh=pos->Property()->Get_Context_Switch_Overhead();
    float scoh=pos->Property()->Get_Scheduling_Overhead();
    float tioh=pos->Property()->Get_Timer_Overhead();
    float tires=pos->Property()->Get_Timer_Resolution();
    
    for(ti=pproc->Tasks_Begin();ti!=pproc->Tasks_End();ti++)
    {
        TaskType* pt=*ti;

        u+=(pt->Property()->Get_Exec_Time()+scoh)/pt->Property()->Get_Global_Period();
    }
    u+=(tires/1000)*tioh;

    if(u>1.00) return false;
    else return true;
}

static void populate_mp_sp (Processor *pproc)
{
    list<TaskType*>::iterator ti1,ti2;

    /* Clear all sets */

    for(ti1=pproc->Tasks_Begin();ti1!=pproc->Tasks_End();ti1++)
    {
        TaskType* pt1=*ti1;

        pt1->Property()->Save_MP();
        pt1->Property()->MP()->clear();
        if(pt1->Property()->Get_Global_Root()==pt1)
        {
            pt1->Property()->Save_SP();
            pt1->Property()->SP()->clear();
            pt1->Property()->LP()->clear();
        }
    }

    for(ti1=pproc->Tasks_Begin();ti1!=pproc->Tasks_End();ti1++)
    {
        TaskType* pt1=*ti1;
        bool first_task=(pt1->Property()->Get_Global_Root()==pt1);

        for(ti2=pproc->Tasks_Begin();ti2!=pproc->Tasks_End();ti2++)
        {
            TaskType* pt2=*ti2;

            if(is_mp(pt2, pt1))
                pt1->Property()->MP()->insert(pt2);
            if((first_task)&&is_sp(pt2,pt1))
                pt1->Property()->SP()->insert(pt2);
            if((first_task)&&is_lp(pt2,pt1))
                pt1->Property()->LP()->insert(pt2);
        }
    }
    
    for(ti1=pproc->Tasks_Begin();ti1!=pproc->Tasks_End();ti1++)
    {
        TaskType* pt1=*ti1;
        bool first_task=(pt1->Property()->Get_Global_Root()==pt1);

        if(first_task)
        {
            float btime=get_blocking_time(pt1);
            pt1->Property()->Set_Block_Time(btime);
        }
    }
}

static void populate_mp_sp (Processor *pproc, TaskType* pt)
{
    list<TaskType*>::iterator ti1,ti2;
    list<TaskType*>* topol=pt->Property()->Topo();
    TaskType* pfirst;

    ti1=topol->begin();
    pfirst=*ti1;
    
    /* Clear all sets */

    pfirst->Property()->Canonize();

    for(ti1=topol->begin();ti1!=topol->end();ti1++)
    {
        TaskType* pt1=*ti1;

        pt1->Property()->Save_MP();
        pt1->Property()->MP()->clear();
        if(pt1->Property()->Get_Global_Root()==pt1)
        {
            pt1->Property()->Save_SP();
            pt1->Property()->SP()->clear();
            pt1->Property()->LP()->clear();
        }
    }

    for(ti1=topol->begin();ti1!=topol->end();ti1++)
    {
        TaskType* pt1=*ti1;
        bool first_task=(pt1->Property()->Get_Global_Root()==pt1);

        for(ti2=pproc->Tasks_Begin();ti2!=pproc->Tasks_End();ti2++)
        {
            TaskType* pt2=*ti2;

            if(is_mp(pt2, pt1))
                pt1->Property()->MP()->insert(pt2);
            if((first_task)&&is_sp(pt2,pt1))
                pt1->Property()->SP()->insert(pt2);
            if((first_task)&&is_lp(pt2,pt1))
                pt1->Property()->LP()->insert(pt2);
        }
    }

    float btime=get_blocking_time(pfirst);
    pfirst->Property()->Set_Block_Time(btime);
}

static bool is_mp(TaskType* pt1, TaskType* pt2)
{
    if(pt1->Property()->Get_Global_Root()!=pt1) return false;
    if(pt1==pt2) return false;

    if(pt1->Property()->Get_Min_Priority()>pt2->Property()->Get_Priority())
        return true;
    else
        return false;
}

static bool is_sp(TaskType* pt1, TaskType* pt2)
{
    if(pt1->Property()->Get_Global_Root()!=pt1) return false;
    if(pt1==pt2) return false;

    int ref_pr=pt2->Property()->Get_Priority();
    int min_pr=pt1->Property()->Get_Min_Priority();
    int last_pr=pt1->Property()->Get_Last_Priority();
    int first_pr=pt1->Property()->Get_First_Priority();

    if(min_pr<=ref_pr)
    {
        if(last_pr>=ref_pr)
        {
            if(first_pr>=ref_pr) return true;
            else return false;
        }
        else
        {
            if(first_pr>=ref_pr) return true;
            else return false;
        }
    }
    else return false;
}

static bool is_lp(TaskType* pt1, TaskType* pt2)
{
    if(pt1->Property()->Get_Global_Root()!=pt1) return false;
    if(pt1==pt2) return false;

    int ref_pr=pt2->Property()->Get_Priority();
    int min_pr=pt1->Property()->Get_Min_Priority();
    int last_pr=pt1->Property()->Get_Last_Priority();
    int first_pr=pt1->Property()->Get_First_Priority();

    if(min_pr<=ref_pr)
    {
        if(last_pr>=ref_pr)
        {
            if(first_pr>=ref_pr) return false;
            else return true;
        }
        else
        {
            if(first_pr>=ref_pr) return false;
            else return true;
        }
    }
    else return false;
}
    
static float sum_h_segments (TaskType* pt1, TaskType* pt)
{
    list<TaskType*>::iterator ti;
    list<TaskType*>* topol=pt1->Property()->Topo();

    float sum=0;
    for(ti=topol->begin();ti!=topol->end();ti++)
    {
        TaskType* pt2=*ti;

        if(pt2->Property()->Get_Priority()>=pt->Property()->Get_Priority())
            sum+=pt2->Property()->Get_Exec_Time();
    }
    return sum;
}

static float get_blocking_time (TaskType* pt)
{
    set<TaskType*>::iterator ti,max_it;
    set<TaskType*>* pSP=pt->Property()->SP();
    set<TaskType*>* pLP=pt->Property()->LP();
    
    float b=0;
    for(ti=pLP->begin();ti!=pLP->end();ti++)
    {
        TaskType* pt1=*ti;

        float bt=sum_h_segments(pt1,pt);
        if(bt>b) b=bt;
    }
    
    float max_val=-9999999;
    float U_x,V_x,W_x;
    for(ti=pSP->begin();ti!=pSP->end();ti++)
    {
        TaskType* pt1=*ti;
        float U=sum_h_inner_segments(pt1,pt);
        float V=sum_h_final_segments(pt1,pt);
        float W=sum_h_initial_segments(pt1,pt);

        float val=max(U-W-b,V-b);
        if(val>max_val)
        {
            max_val=val;
            max_it=ti;
            U_x=U;V_x=V;W_x=W;
        }
    }
    if(max_val<0) return b;
    else
    {
        if(U_x-W_x>V_x)
        {
            pSP->erase(max_it);
            return U_x;
        }
        else
            return V_x;
    }
}
    
static float find_length_busy_period (TaskType* pt)
{
    set<TaskType*>::iterator ti;
    set<TaskType*>* pmp=pt->Property()->MP();
    set<TaskType*>* psp=pt->Property()->SP();
    float pn,p;
    
    /******************* OverHeads *******************/
    Processor* pp=pt->Property()->Get_Proc();
    OS* pos=pp->Get_OS();
    float csoh=pos->Property()->Get_Context_Switch_Overhead();
    float scoh=pos->Property()->Get_Scheduling_Overhead();
    float tioh=pos->Property()->Get_Timer_Overhead();
    float tires=pos->Property()->Get_Timer_Resolution();
    
    float block_time=pt->Property()->Get_Block_Time();
    float exe_time=pt->Property()->Get_Global_Exe_Time();
    p=pt->Property()->Get_Exec_Time();pn=0;
    while(pn!=p)
    {
        float inter=0;
        float sp=0;
        
        pn=p;
        for(ti=pmp->begin();ti!=pmp->end();ti++)
        {
            TaskType* pt1=*ti;
            
            inter+=(ceil(pn/pt1->Property()->Get_Global_Period())*
                    (pt1->Property()->Get_Global_Exe_Time()+2*csoh+scoh));
        }

        for(ti=psp->begin();ti!=psp->end();ti++)
        {
            TaskType* pt1=*ti;

            sp+=(sum_h_initial_segments(pt1,pt)+scoh);
        }

        p=block_time+inter+sp+ceil(pn/pt->Property()->Get_Global_Period())*
            (exe_time+scoh)+(ceil(pn/(1000/tires))*tioh);
    }
    return p;
}

static float find_response_time_first_task (TaskType* pt, int k)
{
    set<TaskType*>::iterator ti;
    set<TaskType*>* pmp=pt->Property()->MP();
    set<TaskType*>* psp=pt->Property()->SP();
    float pn, p;

    /******************* OverHeads *******************/
    Processor* pp=pt->Property()->Get_Proc();
    OS* pos=pp->Get_OS();
    float csoh=pos->Property()->Get_Context_Switch_Overhead();
    float scoh=pos->Property()->Get_Scheduling_Overhead();
    float tioh=pos->Property()->Get_Timer_Overhead();
    float tires=pos->Property()->Get_Timer_Resolution();

    float block_time=pt->Property()->Get_Block_Time();
    float exe_time=pt->Property()->Get_Global_Exe_Time();
    p=pt->Property()->Get_Exec_Time();pn=0;
    while(pn!=p)
    {
        float inter=0;
        float sp=0;

        pn=p;
        for(ti=pmp->begin();ti!=pmp->end();ti++)
        {
            TaskType* pt1=*ti;
            
            inter+=(ceil(pn/pt1->Property()->Get_Global_Period())*
                    (pt1->Property()->Get_Global_Exe_Time()+2*csoh+scoh));
        }
        for(ti=psp->begin();ti!=psp->end();ti++)
        {
            TaskType* pt1=*ti;

            sp+=(sum_h_initial_segments(pt1,pt)+scoh);
        }

        p=block_time+inter+(k-1)*(exe_time+scoh)+pt->Property()->Get_Exec_Time()+
                (ceil(pn/(1000/tires))*tioh)+sp;
    }
    return p;
}

static float find_response_time_sub_task(TaskType* pt, float rel)
{
    set<TaskType*>::iterator ti;
    set<TaskType*>* pmp=pt->Property()->MP();
    set<TaskType*> sp;
	
	get_SP(sp,pt);
    float pn, p;

    /******************* OverHeads *******************/
    Processor* pp=pt->Property()->Get_Proc();
    OS* pos=pp->Get_OS();
    float csoh=pos->Property()->Get_Context_Switch_Overhead();
    float scoh=pos->Property()->Get_Scheduling_Overhead();
    float tioh=pos->Property()->Get_Timer_Overhead();
    float tires=pos->Property()->Get_Timer_Resolution();
    
    p=rel+pt->Property()->Get_Exec_Time();pn=0;
    while(pn!=p)
    {
        float inter=0;
        float sp_val=0;

        pn=p;
        for(ti=pmp->begin();ti!=pmp->end();ti++)
        {
            TaskType* pt1=*ti;
            
            inter+=(
                     (ceil(pn/pt1->Property()->Get_Global_Period()))-
                     (ceil(rel/pt1->Property()->Get_Global_Period()))
                     )*
                    (pt1->Property()->Get_Global_Exe_Time()+2*csoh+scoh);
        }
        for(ti=sp.begin();ti!=sp.end();ti++)
        {
            TaskType* pt1=*ti;

            if((
                        (ceil(pn/pt1->Property()->Get_Global_Period()))-
                        (ceil(rel/pt1->Property()->Get_Global_Period()))
               )>=1)      sp_val+=(sum_h_initial_segments(pt1,pt)+scoh);
        }

        p=scoh+rel+inter+pt->Property()->Get_Exec_Time()+sp_val;
    }
    return p;
}

static void find_response_time (TaskType* pt)
{
    list<TaskType*>::iterator ti;
    list<TaskType*>* tl;
    float busy_period;
    int num_jobs,i;
    
    busy_period=find_length_busy_period(pt);
    num_jobs=(int)ceil(busy_period/pt->Property()->Get_Global_Period());
    
    tl=pt->Property()->Topo();
    for(ti=tl->begin();ti!=tl->end();ti++)
    {
        TaskType* pt1=*ti;

        pt1->Property()->Set_Response_Time(0);
    }

    for(i=0;i<num_jobs;i++)
    {
        float start_time=i*pt->Property()->Get_Global_Period();
        float resp_time=find_response_time_first_task(pt,i+1);

        if((resp_time-start_time)>pt->Property()->Get_Response_Time())
            pt->Property()->Set_Response_Time(resp_time-start_time);

        float release_time=resp_time;
        
        for(ti=tl->begin(),ti++;ti!=tl->end();ti++)
        {
            TaskType* pt1=*ti;
    
            float rt=find_response_time_sub_task (pt1, release_time);
            if((rt-start_time)>pt1->Property()->Get_Response_Time())
                pt1->Property()->Set_Response_Time(rt-start_time);
            release_time=rt;
        }
    }
}

static float sum_h_inner_segments(TaskType* pt1, TaskType* pt)
{
    list<TaskType*>::iterator ti;
    list<TaskType*>* topol=pt1->Property()->Topo();
    float max=0,innerlen=0;
    int state=0; // 0 - outside H, 1 - entering H, 2 - in H
    int ref_priority=pt->Property()->Get_Priority();

    ti=topol->begin();
    for(ti=topol->begin();ti!=topol->end();ti++)
    {
        TaskType* pt2=*ti;

        if(pt2->Property()->Get_Priority()<ref_priority)
        {
            switch(state)
            {
                case 0: state=1; break;
                case 1: break;
                case 2: 
                        if(innerlen>max) max=innerlen;
                        innerlen=0;
                        state=0;
                        break;
            }
        }
        else
        {
            switch(state)
            {
                case 0: break;
                case 1: state=2; break;
                case 2:
                        innerlen+=pt2->Property()->Get_Exec_Time();
                        break;
            }
        }
    }
    return max;
}
    
static float sum_h_initial_segments(TaskType* pt1, TaskType* pt)
{
    list<TaskType*>::iterator ti;
    list<TaskType*>* topol=pt1->Property()->Topo();
    float sum=0;
    int ref_priority=pt->Property()->Get_Priority();

    for(ti=topol->begin();ti!=topol->end();ti++)
    {
        TaskType* pt2=*ti;

        if(pt2->Property()->Get_Priority()<ref_priority)
            break;
        else
            sum+=pt2->Property()->Get_Exec_Time();
    }
    return sum;
}

static float sum_h_final_segments(TaskType* pt1, TaskType* pt)
{
    list<TaskType*>::reverse_iterator ti;
    list<TaskType*>* topol=pt1->Property()->Topo();
    float sum=0;
    int ref_priority=pt->Property()->Get_Priority();

    for(ti=topol->rbegin();ti!=topol->rend();ti++)
    {
        TaskType* pt2=*ti;

        if(pt2->Property()->Get_Priority()<ref_priority)
            break;
        else
            sum+=pt2->Property()->Get_Exec_Time();
    }
    return sum;
}

static void get_SP (set<TaskType*>& lt, TaskType* pt)
{
    set<TaskType*>* mp=pt->Property()->MP();
    set<TaskType*>* prev_mp=pt->Property()->Get_Predecessor()->Property()->MP();
    set<TaskType*>::iterator ti,si;

    for(ti=prev_mp->begin();ti!=prev_mp->end();ti++)
    {
        TaskType* pt1=*ti;

        si=mp->find(pt1);
        if(si==mp->end())
        {
            if(is_sp(pt1,pt)) lt.insert(pt1);
        }
    }
}
